"""
Various hooks for extending Splitgraph, including:

  * External object handlers (:mod:`splitgraph.hooks.external_objects`) allowing to download/upload objects
    to locations other than the remote Splitgraph engine.
  * Data sources (:mod:`splitgraph.hooks.data_sources`) that allow to add data to Splitgraph, e.g.
   using the Postgres engine's FDW interface to mount other external databases on the engine.
  * Splitfile commands (:mod:`splitgraph.hooks.splitfile_commands`) to define custom data transformation steps
    compatible with the Splitfile framework.
"""



"""
Functions for executing Splitfiles.
"""

import json
import logging
from hashlib import sha256
from importlib import import_module
from random import getrandbits
from typing import Callable, Dict, List, Optional, Tuple, cast

from parsimonious.nodes import Node

from splitgraph.config import CONFIG
from splitgraph.config.config import get_all_in_section, get_singleton
from splitgraph.core.engine import lookup_repository, repository_exists
from splitgraph.core.image import Image
from splitgraph.core.repository import Repository, clone
from splitgraph.core.sql import prepare_splitfile_sql, validate_import_sql
from splitgraph.engine import get_engine
from splitgraph.exceptions import ImageNotFoundError, SplitfileError
from splitgraph.hooks.mount_handlers import mount

from ..core.image_mounting import ImageMapper, _get_local_image_for_import
from ..core.output import Color, conn_string_to_dict, pluralise, truncate_line
from ..core.types import ProvenanceLine
from ._parsing import (
    extract_all_table_aliases,
    extract_nodes,
    get_first_or_none,
    parse_commands,
    parse_custom_command,
    parse_image_spec,
)


def _combine_hashes(hashes: List[str]) -> str:
    return sha256("".join(hashes).encode("ascii")).hexdigest()


def _checkout_or_calculate_layer(output: Repository, image_hash: str, calc_func: Callable) -> None:
    # Future optimization here: don't actually check the layer out if it exists -- only do it at Splitfile execution
    # end or when a command needs it.

    # Have we already calculated this hash?
    try:
        output.images.by_hash(image_hash).checkout()
        logging.info(" ---> Using cache")
    except ImageNotFoundError:
        try:
            calc_func()
        except Exception:
            # Some of the Splitfile commands have to use a commit
            # (e.g. IMPORT uses Image.query_schema context manager that mounts the source
            # image via LQ and commits), so a rollback won't delete this new image
            # that we might have created.
            output.images.delete([image_hash])
            raise
    logging.info(" ---> %s" % image_hash[:12])


def execute_commands(
    commands: str,
    params: Optional[Dict[str, str]] = None,
    output: Optional[Repository] = None,
    output_base: str = "0" * 32,
) -> None:
    """
    Executes a series of Splitfile commands.

    :param commands: A string with the raw Splitfile.
    :param params: A dictionary of parameters to be applied to the Splitfile (`${PARAM}` is replaced with the specified
        parameter value).
    :param output: Output repository to execute the Splitfile against.
    :param output_base: If not None, a revision that gets checked out for all Splitfile actions to be committed
        on top of it.
    """
    if params is None:
        params = {}
    # Use a random target schema if unspecified.
    output = output or Repository.from_schema("output_%0.2x" % getrandbits(16))

    # Don't initialize the output until a command writing to it asks us to
    # (otherwise we might have a FROM ... AS output_name change it).
    repo_created = False
    output_base_checked_out = False

    def _initialize_output(output):
        if not repository_exists(output):
            nonlocal repo_created
            output.init()
            repo_created = True

        nonlocal output_base_checked_out
        if (
            not output_base_checked_out
            and output
            and repository_exists(output)
            and output_base is not None
        ):
            # If we're running commands and the Splitfile doesn't start with FROM to set the base
            # image, check out the output_base (delay it so that checking out 000... and then
            # running FROM doesn't do redundant work).
            checkout_if_changed(output, output_base)
            output_base_checked_out = True

    node_list = parse_commands(commands, params=params)

    # Record the internal structure of commands used to create the final image.
    provenance: List[ProvenanceLine] = []

    try:
        for i, node in enumerate(node_list):
            logging.info(
                Color.BOLD
                + "\nStep %d/%d : %s" % (i + 1, len(node_list), truncate_line(node.text, length=60))
                + Color.END
            )
            if node.expr_name == "from":
                output, maybe_provenance_line = _execute_from(node, output)
                output_base_checked_out = True
                if maybe_provenance_line:
                    provenance.append(maybe_provenance_line)

            elif node.expr_name == "import":
                _initialize_output(output)
                provenance_line = _execute_import(node, output)
                provenance.append(provenance_line)

            elif node.expr_name == "sql" or node.expr_name == "sql_file":
                _initialize_output(output)
                provenance_line = _execute_sql(node, output)
                provenance.append(provenance_line)

            elif node.expr_name == "custom":
                _initialize_output(output)
                provenance_line = _execute_custom(node, output)
                provenance.append(provenance_line)

        final_image = output.head_strict
        final_image.set_provenance(provenance)
        get_engine().commit()
        logging.info("Successfully built %s:%s." % (str(output), final_image.image_hash[:12]))

    except Exception:
        if repo_created and len(output.images()) == 1:
            # As a corner case, if we created a repository and there's been
            # a failure running the Splitfile (on the first command), we delete the dummy
            # 0000... image and the rest of the repository as part of cleanup.
            output.delete()
        get_engine().rollback()
        raise


def checkout_if_changed(repository: Repository, image_hash: str) -> None:
    if (
        repository.head is None
        or (repository.head.image_hash != image_hash)
        or repository.has_pending_changes()
    ):
        repository.images.by_hash(image_hash).checkout()
    else:
        logging.info(
            "Skipping checkout of %s as %s has it checked out " "and there have been no changes",
            image_hash,
            repository,
        )


def _execute_sql(node: Node, output: Repository) -> ProvenanceLine:
    # Calculate the hash of the layer we are trying to create.
    # Since we handle the "input" hashing in the import step, we don't need to care about the sources here.
    # Later on, we could enhance the caching and base the hash of the command on the hashes of objects that
    # definitely go there as sources.
    if node.expr_name == "sql_file":
        node_contents = extract_nodes(node, ["non_newline"])[0].text
        logging.info("Loading the SQL commands from %s" % node_contents)
        with open(node_contents, "r") as file:
            # Possibly use a different method to calculate the image hash for commands originating from
            # SQL files instead?
            # Don't "canonicalize" it here to get rid of whitespace, just hash the whole file.
            sql_command = file.read()
    else:
        # Support singleline and multiple SQL commands between braces
        nodes = extract_nodes(node, ["non_newline"])
        if not nodes:
            nodes = extract_nodes(node, ["non_curly_brace"])
        sql_command = nodes[0].text.replace("\\{", "{").replace("\\}", "}").replace("\\\\", "\\")

    # Prepare the SQL by rewriting outside image references into LQ schemas
    # Canonicalize the SQL command by formatting it.
    image_mapper = ImageMapper(output.object_engine)
    sql_rewritten, sql_canonical = prepare_splitfile_sql(sql_command, image_mapper)

    output_head = output.head_strict.image_hash
    target_hash = _combine_hashes([output_head, sha256(sql_canonical.encode("utf-8")).hexdigest()])

    def _calc():
        logging.info("Executing SQL...")
        try:
            image_mapper.setup_lq_mounts()
            output.object_engine.commit()
            sql_to_run = get_singleton(CONFIG, "SG_LQ_TUNING") + sql_rewritten
            logging.debug("Running rewritten SQL %s against %s", sql_to_run, output)
            output.run_sql(sql_to_run)
        finally:
            image_mapper.teardown_lq_mounts()
        output.commit(target_hash, comment=sql_command)

    _checkout_or_calculate_layer(output, target_hash, _calc)
    provenance: ProvenanceLine = {"type": "SQL", "sql": sql_canonical}
    provenance.update(image_mapper.get_provenance_data())
    return provenance


def _execute_from(node: Node, output: Repository) -> Tuple[Repository, Optional[ProvenanceLine]]:
    interesting_nodes = extract_nodes(node, ["repo_source", "repository"])
    repo_source = get_first_or_none(interesting_nodes, "repo_source")
    output_node = get_first_or_none(interesting_nodes, "repository")
    provenance: Optional[ProvenanceLine] = None

    if output_node:
        # AS (output) detected, change the current output repository to it.
        output = Repository.from_schema(output_node.match.group(0))
        logging.info("Changed output repository to %s" % str(output))

        # NB this destroys all data in the case where we ran some commands in the Splitfile and then
        # did FROM (...) without AS repository
        if repository_exists(output):
            logging.info("Clearing all output from %s" % str(output))
            output.delete()
    if not repository_exists(output):
        output.init()
    if repo_source:
        repository, tag_or_hash = parse_image_spec(repo_source)
        source_repo = lookup_repository(repository.to_schema(), include_local=True)

        if source_repo.engine.name == "LOCAL" and source_repo.upstream:
            # For local repositories, make sure to update them if they've an upstream
            source_repo.pull(single_image=tag_or_hash)

        # Get the target image hash from the source repo: otherwise, if the tag is, say, 'latest' and
        # the output has just had the base commit (000...) created in it, that commit will be the latest.
        clone(source_repo, local_repository=output, download_all=False)
        source_hash = source_repo.images[tag_or_hash].image_hash
        checkout_if_changed(output, source_hash)
        provenance = {
            "type": "FROM",
            "source_namespace": source_repo.namespace,
            "source": source_repo.repository,
            "source_hash": source_hash,
        }
    else:
        # FROM EMPTY AS repository -- initializes an empty repository (say to create a table or import
        # the results of a previous stage in a multistage build.
        # In this case, if AS repository has been specified, it's already been initialized. If not, this command
        # literally does nothing
        if not output_node:
            raise SplitfileError("FROM EMPTY without AS (repository) does nothing!")
    return output, provenance


def _execute_import(node: Node, output: Repository) -> ProvenanceLine:
    interesting_nodes = extract_nodes(node, ["repo_source", "mount_source", "tables"])
    table_names, table_aliases, table_queries = extract_all_table_aliases(interesting_nodes[-1])
    if interesting_nodes[0].expr_name == "repo_source":
        # Import from a repository (local or remote)
        repository, tag_or_hash = parse_image_spec(interesting_nodes[0])
        return _execute_repo_import(
            repository, table_names, tag_or_hash, output, table_aliases, table_queries
        )
    else:
        # Extract the identifier (FDW name), the connection string and the FDW params (JSON-encoded, everything
        # between the single quotes).
        mount_nodes = extract_nodes(
            interesting_nodes[0], ["identifier", "no_db_conn_string", "non_single_quote"]
        )
        fdw_name = mount_nodes[0].match.group(0)
        conn_string = mount_nodes[1].match
        fdw_params = mount_nodes[2].match.group(0).replace("\\'", "'")  # Unescape the single quote

        return _execute_db_import(
            conn_string, fdw_name, fdw_params, table_names, output, table_aliases, table_queries
        )


def _execute_db_import(
    conn_string, fdw_name, fdw_params, table_names, target_mountpoint, table_aliases, table_queries
) -> ProvenanceLine:
    tmp_mountpoint = Repository.from_schema(fdw_name + "_tmp_staging")
    try:
        handler_kwargs = json.loads(fdw_params)
        handler_kwargs.update(conn_string_to_dict(conn_string.group() if conn_string else None))
        mount(tmp_mountpoint.to_schema(), fdw_name, handler_kwargs)
        # The foreign database is a moving target, so the new image hash is random.
        # Maybe in the future, when the object hash is a function of its contents, we can be smarter here...
        target_hash = "{:064x}".format(getrandbits(256))
        target_mountpoint.import_tables(
            table_aliases,
            tmp_mountpoint,
            table_names,
            target_hash=target_hash,
            foreign_tables=True,
            table_queries=table_queries,
        )
        return {"type": "MOUNT"}
    finally:
        tmp_mountpoint.delete()


def prevalidate_imports(table_names: List[str], table_queries: List[bool]) -> List[str]:
    return [validate_import_sql(t) if q else t for t, q in zip(table_names, table_queries)]


def _execute_repo_import(
    repository: Repository,
    table_names: List[str],
    tag_or_hash: str,
    target_repository: Repository,
    table_aliases: List[str],
    table_queries: List[bool],
) -> ProvenanceLine:
    # Don't use the actual routine here as we want more control: clone the remote repo in order to turn
    # the tag into an actual hash
    local_image, repo_is_temporary = _get_local_image_for_import(tag_or_hash, repository)
    source_hash = local_image.image_hash
    source_mountpoint = local_image.repository
    try:

        # Perform validation here rather than in the import routine. This is to
        # canonicalize SQL that goes into provenance data so that we don't get cache misses
        # from formatting changes to SQL statements.
        table_names_canonical = prevalidate_imports(table_names, table_queries)

        # Calculate the hash of the new layer by combining the hash of the previous layer,
        # the hash of the source and all the table names/aliases getting imported.
        # This can be made more granular later by using, say, the object IDs of the tables
        # that are getting imported (so that if there's a new commit with some of the same objects,
        # we don't invalidate the downstream).
        # If table_names actually contains queries that generate data from tables, we can still use
        # it for hashing: we assume that the queries are deterministic, so if the query is changed,
        # the whole layer is invalidated.
        output_head = target_repository.head_strict.image_hash
        target_hash = _combine_hashes(
            [output_head, source_hash]
            + [sha256(n.encode("utf-8")).hexdigest() for n in table_names_canonical + table_aliases]
        )

        def _calc():
            logging.info(
                "Importing %s from %s:%s into %s",
                pluralise("table", len(table_names)),
                str(repository),
                source_hash[:12],
                str(target_repository),
            )
            target_repository.import_tables(
                table_aliases,
                source_mountpoint,
                table_names,
                image_hash=source_hash,
                target_hash=target_hash,
                table_queries=table_queries,
                skip_validation=True,
            )

        _checkout_or_calculate_layer(target_repository, target_hash, _calc)
        return {
            "type": "IMPORT",
            "source_namespace": repository.namespace,
            "source": repository.repository,
            "source_hash": source_hash,
            "tables": table_names_canonical,
            "table_aliases": table_aliases,
            "table_queries": table_queries,
        }
    finally:
        if repo_is_temporary:
            source_mountpoint.delete()


def _execute_custom(node: Node, output: Repository) -> ProvenanceLine:
    assert output.head is not None
    command, args = parse_custom_command(node)

    # Locate the command in the config file and instantiate it.
    cmd_fq_class: str = cast(str, get_all_in_section(CONFIG, "commands").get(command))
    if not cmd_fq_class:
        raise SplitfileError(
            "Custom command {0} not found in the config! Make sure you add an entry to your"
            " config like so:\n  [commands]  \n{0}=path.to.command.Class".format(command)
        )

    assert isinstance(cmd_fq_class, str)
    index = cmd_fq_class.rindex(".")
    try:
        cmd_class = getattr(import_module(cmd_fq_class[:index]), cmd_fq_class[index + 1 :])
    except AttributeError as e:
        raise SplitfileError("Error loading custom command {0}".format(command)) from e
    except ImportError as e:
        raise SplitfileError("Error loading custom command {0}".format(command)) from e

    get_engine().run_sql("SET search_path TO %s", (output.to_schema(),))
    command = cmd_class()

    # Pre-flight check: get the new command hash and see if we can short-circuit and just check the image out.
    command_hash = command.calc_hash(repository=output, args=args)
    output_head = output.head.image_hash

    if command_hash is not None:
        image_hash = _combine_hashes([output_head, command_hash])
        try:
            output.images.by_hash(image_hash).checkout()
            logging.info(" ---> Using cache")
            return {"type": "CUSTOM"}
        except ImageNotFoundError:
            pass

    logging.info(" Executing custom command...")
    exec_hash = command.execute(repository=output, args=args)
    command_hash = command_hash or exec_hash or "{:064x}".format(getrandbits(256))

    image_hash = _combine_hashes([output_head, command_hash])
    logging.info(" ---> %s" % image_hash[:12])

    # Check just in case if the new hash produced by the command already exists.
    try:
        output.images.by_hash(image_hash).checkout()
    except ImageNotFoundError:
        # Full command as a commit comment
        output.commit(image_hash, comment=node.text)
    return {"type": "CUSTOM"}


def rebuild_image(image: Image, source_replacement: Dict[Repository, str]) -> None:
    """
    Recreates the Splitfile used to create a given image and reruns it, replacing its dependencies with a different
    set of versions.

    :param image: Image object
    :param source_replacement: A map that specifies replacement images/tags for repositories that the image depends on
    """
    splitfile_commands = image.to_splitfile(
        ignore_irreproducible=False, source_replacement=source_replacement
    )
    # Params are supposed to be stored in the commands already (baked in) -- what if there's sensitive data there?
    execute_commands("\n".join(splitfile_commands), output=image.repository)



"""
Miscellaneous image management sgr commands.
"""
import atexit
import contextlib
import os
import platform
import shutil
import stat
import subprocess
import sys
import uuid
from pathlib import Path
from typing import Optional, Tuple

import click
from packaging.version import Version

from splitgraph.__version__ import __version__
from splitgraph.exceptions import CheckoutError

from .common import ImageType, RepositoryType, remote_switch_option
from .engine import list_engines


@click.command(name="rm")
@click.argument("image_spec", type=ImageType(default=None))
@remote_switch_option()
@click.option(
    "-y", "--yes", help="Agree to deletion without confirmation", is_flag=True, default=False
)
def rm_c(image_spec, yes):
    """
    Delete schemas, repositories or images.

    If the target of this command is a Postgres schema, this performs DROP SCHEMA CASCADE.

    If the target of this command is a Splitgraph repository, this deletes the repository and all of its history.

    If the target of this command is an image, this deletes the image and all of its children.

    In any case, this command will ask for confirmation of the deletion, unless ``-y`` is passed. If ``-r``
    (``--remote``), is passed, this will perform deletion on a remote Splitgraph engine (registered in the config)
    instead, assuming the user has write access to the remote repository.

    This does not delete any physical objects that the deleted repository/images depend on:
    use ``sgr cleanup`` to do that.

    Examples:

    ``sgr rm temporary_schema``

    Deletes ``temporary_schema`` from the local engine.

    ``sgr rm --remote data.splitgraph.com username/repo``

    Deletes ``username/repo`` from the Splitgraph registry.

    ``sgr rm -y username/repo:old_branch``

    Deletes the image pointed to by ``old_branch`` as well as all of its children (images created by a commit based
    on this image), as well as all of the tags that point to now deleted images, without asking for confirmation.
    Note this will not delete images that import tables from the deleted images via Splitfiles or indeed the
    physical objects containing the actual tables.
    """
    from splitgraph.core.engine import repository_exists
    from splitgraph.core.repository import Repository
    from splitgraph.engine import get_engine

    engine = get_engine()

    repository, image = image_spec
    repository = Repository.from_template(repository, engine=engine)
    if not image:
        click.echo(
            ("Repository" if repository_exists(repository) else "Postgres schema")
            + " %s will be deleted." % repository.to_schema()
        )
        if not yes:
            click.confirm("Continue? ", abort=True)

        # Don't try to "uncheckout" repositories on the registry/other remote engines
        repository.delete(uncheckout=engine.name == "LOCAL")
        repository.commit_engines()
    else:
        image = repository.images[image]
        images_to_delete = repository.images.get_all_child_images(image.image_hash)
        tags_to_delete = [t for i, t in repository.get_all_hashes_tags() if i in images_to_delete]

        click.echo("Images to be deleted:")
        click.echo("\n".join(sorted(images_to_delete)))
        click.echo("Total: %d" % len(images_to_delete))

        click.echo("\nTags to be deleted:")
        click.echo("\n".join(sorted(tags_to_delete)))
        click.echo("Total: %d" % len(tags_to_delete))

        if "HEAD" in tags_to_delete:
            # If we're deleting an image that we currently have checked out,
            # we need to make sure the rest of the metadata (e.g. current state of the audit table) is consistent,
            # it's better to disallow these deletions completely.
            raise CheckoutError(
                "Deletion will affect a checked-out image! Check out a different branch "
                "or do sgr checkout -u %s!" % repository.to_schema()
            )
        if not yes:
            click.confirm("Continue? ", abort=True)

        repository.images.delete(images_to_delete)
        repository.commit_engines()
        click.echo("Success.")


@click.command(name="prune")
@click.argument("repository", type=RepositoryType(exists=True))
@remote_switch_option()
@click.option(
    "-y", "--yes", help="Agree to deletion without confirmation", is_flag=True, default=False
)
def prune_c(repository, yes):
    """
    Cleanup dangling images from a repository.

    This includes images not pointed to by any tags (or checked out) and those that aren't required by any of
    such images.

    Will ask for confirmation of the deletion, unless ``-y ``is passed. If ``-r`` (``--remote``) is
    passed, this will perform deletion on a remote Splitgraph engine (registered in the config) instead, assuming
    the user has write access to the remote repository.

    This does not delete any physical objects that the deleted repository/images depend on:
    use ``sgr cleanup`` to do that.
    """
    from splitgraph.core.repository import Repository
    from splitgraph.engine import get_engine

    repository = Repository.from_template(repository, engine=get_engine())

    all_images = {image.image_hash for image in repository.images()}
    all_tagged_images = {i for i, t in repository.get_all_hashes_tags() if i}
    dangling_images = all_images.difference(
        repository.images.get_all_parent_images(all_tagged_images)
    )

    if not dangling_images:
        click.echo("Nothing to do.")
        return

    click.echo("Images to be deleted:")
    click.echo("\n".join(sorted(dangling_images)))
    click.echo("Total: %d" % len(dangling_images))

    if not yes:
        click.confirm("Continue? ", abort=True)

    repository.images.delete(dangling_images)
    repository.commit_engines()
    click.echo("Success.")


@click.command(name="init")
@click.argument("repository", type=RepositoryType(), required=False, default=None)
@click.option("--skip-object-handling", default=False, is_flag=True)
def init_c(repository, skip_object_handling):
    """
    Initialize a new repository/engine.

    Examples:

    `sgr init`

    Initializes the current local Splitgraph engine by writing some bookkeeping information.
    This is required for the rest of sgr to work.

    `sgr init --skip-object-handling`

    Initializes a Splitgraph engine without installing audit triggers or object management routines:
    this is useful for engines that aren't intended to be used for image checkouts.

    ``sgr init new/repo``

    Creates a single image with the hash ``00000...`` in ``new/repo``
    """
    from splitgraph.core.engine import init_engine

    if repository:
        if skip_object_handling:
            raise click.BadOptionUsage(
                "--skip-object-handling", "Unsupported when initializing a new repository!"
            )
        repository.init()
        click.echo("Initialized empty repository %s" % str(repository))
    else:
        init_engine(skip_object_handling=skip_object_handling)


@click.command(name="cleanup")
def cleanup_c():
    """
    Prune unneeded objects from the engine.

    This deletes all objects from the cache that aren't required by any local repository.
    """
    from splitgraph.core.object_manager import ObjectManager
    from splitgraph.engine import get_engine

    from ..core.output import pluralise

    deleted = ObjectManager(get_engine()).cleanup()
    click.echo("Deleted %s." % pluralise("object", len(deleted)))


@click.command(name="config")
@click.option(
    "-s",
    "--no-shielding",
    is_flag=True,
    default=False,
    help="If set, doesn't replace sensitive values (like passwords) with asterisks",
)
@click.option(
    "-c",
    "--config-format",
    is_flag=True,
    default=False,
    help="Output configuration in the Splitgraph config file format",
)
@click.option(
    "-n",
    "--conn-string",
    is_flag=True,
    default=False,
    help="Print a libpq connection string to the engine",
)
def config_c(no_shielding, config_format, conn_string):
    """
    Print the current Splitgraph configuration.

    This takes into account the local config file, the default values
    and all overrides specified via environment variables.

    This command can be used to dump the current Splitgraph configuration into a file:

    ```
    sgr config --no-shielding --config-format > .sgconfig
    ```

    ...or save a config file overriding an entry:

    ```
    SG_REPO_LOOKUP=engine1,engine2 sgr config -sc > .sgconfig
    ```

    If `--conn-string` is passed, this prints out a libpq connection string
    that can be used to connect to the default Splitgraph engine with other tools:

    ```
    pgcli $(sgr config -n)
    ```
    """

    from splitgraph.config import CONFIG
    from splitgraph.config.export import serialize_config
    from splitgraph.engine import get_engine
    from splitgraph.engine.postgres.engine import get_conn_str

    if conn_string:
        click.echo(get_conn_str(get_engine().conn_params))
        return

    click.echo(serialize_config(CONFIG, config_format, no_shielding))


@click.command(name="dump")
@click.argument("repository", type=RepositoryType(exists=True))
@click.option(
    "--exclude-object-contents",
    is_flag=True,
    default=False,
    help="Don't dump the commands needed to recreate objects required by the repository.",
)
def dump_c(repository, exclude_object_contents):
    """
    Dump a repository to SQL.
    """
    repository.dump(sys.stdout, exclude_object_contents=exclude_object_contents)


def _eval(command, args):
    # appease PyCharm
    # noinspection PyUnresolvedReferences
    from splitgraph.core.object_manager import ObjectManager
    from splitgraph.core.repository import Repository  # noqa
    from splitgraph.engine import get_engine

    engine = get_engine()
    object_manager = ObjectManager(object_engine=engine, metadata_engine=engine)

    command_locals = locals().copy()
    command_locals.update({k: v for k, v in args})

    # The whole point of this function is to unsafely run Python code from the cmdline,
    # so silence the Bandit warning.
    exec(command, globals(), command_locals)  # nosec


@click.command(name="eval")
@click.option(
    "--i-know-what-im-doing",
    is_flag=True,
    help="Pass this if you're sure that the code you're running "
    "is safe and don't want to be prompted.",
)
@click.argument("command", type=str)
@click.option(
    "-a",
    "--arg",
    multiple=True,
    type=(str, str),
    help="Make extra variables available in the command's namespace",
)
def eval_c(i_know_what_im_doing, command, arg):
    """
    Evaluate a Python snippet using the Splitgraph API.

    This is for advanced users only and should be only used
    if you know what you are doing.

    Normal Python statements are supported and the command is evaluated
    in a namespace where the following is already imported and available:

      * Repository: class that instantiates a Splitgraph repository and makes
        API functions like .commit(), .checkout() etc available.

      * engine: Current local engine

      * object_manager: an instance of ObjectManager that allows
        to get information about objects and manage the object cache.

    \b
    Example:
    ```
    sgr eval 'import json; print(json.dumps(Repository\\
        .from_schema(repo_name)\\
        .images["latest"]\\
        .get_table(table_name)\\
        .table_schema))' \\
    -a repo_name my_repo -a table_name my_table
    ```

    Will dump the schema of table my_table in the most recent image in my_repo in JSON format.

    For more information, see the Splitgraph API reference.
    """
    if not i_know_what_im_doing:
        click.confirm(
            "WARNING: This command might be unsafe and break your Splitgraph \n"
            "installation or harm your machine. This is exactly the same as running \n"
            "untrusted code off of the Internet. In addition, it might rely on undocumented \n"
            "Splitgraph internals that might change in the future. \n\n"
            "Have you checked it and made sure it's not doing anything suspicious?",
            abort=True,
        )

    _eval(command, arg)


def _get_binary_url_for(system, release: str = "latest") -> Tuple[str, str]:
    import requests

    from splitgraph.cloud import get_headers

    if release == "latest":
        endpoint = "https://api.github.com/repos/splitgraph/splitgraph/releases/%s" % release
    else:
        endpoint = "https://api.github.com/repos/splitgraph/splitgraph/releases/tags/%s" % release
    headers = get_headers()
    headers.update({"Accept": "application/vnd.github.v3+json"})
    response = requests.get(endpoint, headers=get_headers())
    if response.status_code == 404:
        raise ValueError("No releases found for tag %s, system %s!" % (release, system))
    response.raise_for_status()

    body = response.json()
    actual_version = body["tag_name"].lstrip("v")
    executable = "sgr-%s-x86_64" % system
    if system == "windows":
        executable += ".exe"
    asset = [a for a in body["assets"] if a["name"] == executable]
    if not asset:
        raise ValueError("No releases found for tag %s, system %s!" % (release, system))
    return actual_version, asset[0]["browser_download_url"]


@click.command(name="upgrade")
@click.option("--skip-engine-upgrade", is_flag=True, help="Only upgrade the client")
@click.option("--path", help="Override the path to download the new binary to.")
@click.option("--force", is_flag=True, help="Reinstall older/same versions.")
@click.argument("version", default="latest")
def upgrade_c(skip_engine_upgrade, path, force, version):
    """Upgrade sgr client and engine.

    This will try to download the most recent stable binary for the current platform
    into the location this binary is running from and then upgrade the default engine.

    This method is only supported for single-binary installs and engines managed
    by `sgr engine`.
    """
    import requests
    from tqdm import tqdm

    from splitgraph.cloud import get_headers
    from splitgraph.config import CONFIG, SG_CMD_ASCII

    # Detect if we're running from a Pyinstaller binary
    if not hasattr(sys, "frozen") and not force and not path:
        raise click.ClickException(
            "Not running from a single binary. Use the tool "
            "you originally used to install Splitgraph (e.g. pip) "
            "to upgrade. To download and run a single binary "
            "for your platform, pass --force and --path."
        )

    system = _get_system_id()

    # Get link to the release
    actual_version, download_url = _get_binary_url_for(
        system, "v" + version if version != "latest" else version
    )

    if version == "latest":
        click.echo("Latest version is %s." % actual_version)

    if Version(actual_version) < Version(__version__) and not force:
        click.echo(
            "Newer or existing version of sgr (%s) is installed, pass --force to reinstall."
            % __version__
        )
        return

    # Download the file

    temp_path, final_path = _get_download_paths(path, download_url)

    # Delete the temporary file at exit if we crash

    def _unlink():
        with contextlib.suppress(FileNotFoundError):
            temp_path.unlink()

    atexit.register(_unlink)

    click.echo("Downloading sgr %s from %s to %s" % (actual_version, download_url, temp_path))

    headers = get_headers()
    response = requests.get(download_url, headers=headers, allow_redirects=True, stream=True)
    response.raise_for_status()

    with tqdm(
        total=int(response.headers["Content-Length"]), unit="B", unit_scale=True, ascii=SG_CMD_ASCII
    ) as pbar:
        with (open(str(temp_path), "wb")) as f:
            for chunk in response.iter_content(chunk_size=1024):
                f.write(chunk)
                pbar.update(len(chunk))

    # Test the new binary
    st = os.stat(str(temp_path))
    os.chmod(str(temp_path), st.st_mode | stat.S_IEXEC)
    subprocess.check_call([str(temp_path), "--version"])

    # Upgrade the default engine
    if not skip_engine_upgrade:
        containers = list_engines(include_all=True, prefix=CONFIG["SG_ENGINE_PREFIX"])
        if containers:
            click.echo("Upgrading the default engine...")
            subprocess.check_call([str(temp_path), "engine", "upgrade"])

    def _finalize():
        # On Windows we can't replace a running executable + probably should keep a backup anyway.
        # We hence rename ourselves to a .old file and ask the user to delete it if needed.
        if final_path.exists():
            backup_path = final_path.with_suffix(final_path.suffix + ".old")
            with contextlib.suppress(FileNotFoundError):
                # shutil.move() breaks on Windows otherwise.
                backup_path.unlink()
            shutil.move(str(final_path), str(backup_path))
            click.echo("Old version has been backed up to %s." % backup_path)

        shutil.move(str(temp_path), str(final_path))
        click.echo("New sgr has been installed at %s." % final_path)

    # Instead of moving the file right now, do it at exit -- that way
    # Pyinstaller won't break with a scary error when cleaning up because
    # the file it's running from has changed.
    atexit.register(_finalize)


def _get_download_paths(path: Optional[str], download_url: str) -> Tuple[Path, Path]:
    if path:
        path = os.path.abspath(path)
        if os.path.isdir(path):
            destdir = path
            file_name = download_url.split("/")[-1]
        else:
            destdir = os.path.dirname(path)
            file_name = os.path.basename(path)
    else:
        destdir = os.path.dirname(sys.executable)
        file_name = os.path.basename(sys.executable)
    final_path = Path(destdir) / file_name
    temp_path = Path(destdir) / uuid.uuid4().hex
    return temp_path, final_path


def _get_system_id():
    system_reported = platform.system()
    try:
        return {"Linux": "linux", "Windows": "windows", "Darwin": "osx"}[system_reported]
    except KeyError:
        raise click.ClickException("Single binary is unsupported on system %s!" % system_reported)



import json
import logging
import os
import subprocess
import tempfile
from abc import ABC, abstractmethod
from contextlib import contextmanager
from io import StringIO
from threading import Thread
from typing import Optional, cast

from splitgraph.core.repository import Repository
from splitgraph.core.types import IntrospectionResult, SyncState, TableInfo, TableParams
from splitgraph.exceptions import DataSourceError
from splitgraph.hooks.data_source.base import (
    SyncableDataSource,
    get_ingestion_state,
    prepare_new_image,
)
from splitgraph.ingestion.common import add_timestamp_tags
from splitgraph.ingestion.singer.common import (
    SingerCatalog,
    SingerConfig,
    SingerState,
    store_ingestion_state,
)
from splitgraph.ingestion.singer.db_sync import (
    get_key_properties,
    get_sg_schema,
    get_table_name,
    run_patched_sync,
    select_breadcrumb,
)


class SingerDataSource(SyncableDataSource, ABC):
    # Some taps (e.g. tap-github) use legacy --properties instead of --catalog
    use_properties = False

    # When True, the tap uses stream["schema"]["selected"] instead of
    #  stream["metadata"][breadcrumb == []]["selected"
    use_legacy_stream_selection = False

    @abstractmethod
    def get_singer_executable(self):
        raise NotImplementedError

    def get_singer_config(self):
        return {**self.params, **self.credentials}

    def _run_singer_discovery(self, config: Optional[SingerConfig] = None) -> SingerCatalog:
        executable = self.get_singer_executable()
        args = [executable]

        with tempfile.TemporaryDirectory() as d:
            self._add_file_arg(config, "config", d, args)

            catalog = subprocess.check_output(args + ["--discover"])

        return cast(SingerCatalog, json.loads(catalog))

    @staticmethod
    def _add_file_arg(var, var_name, dir, args):
        if var:
            path = os.path.join(dir, "%s.json" % var_name)
            with open(path, "w") as f:
                json.dump(var, f)
            args.extend(["--%s" % var_name, path])

    @contextmanager
    def _run_singer(
        self,
        config: Optional[SingerConfig] = None,
        state: Optional[SingerState] = None,
        catalog: Optional[SingerCatalog] = None,
    ):
        executable = self.get_singer_executable()

        args = [executable]

        with tempfile.TemporaryDirectory() as d:
            self._add_file_arg(config, "config", d, args)
            self._add_file_arg(state, "state", d, args)
            self._add_file_arg(catalog, "properties" if self.use_properties else "catalog", d, args)
            logging.debug("Singer catalog: %s", catalog)
            logging.info("Running Singer tap. Arguments: %s", args)
            proc = subprocess.Popen(
                args, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE
            )

            def _emit_output(pipe):
                for line in iter(pipe.readline, b""):
                    logging.info(line.decode("utf-8").rstrip("\n"))

            # Emit stderr as a separate thread.
            t = Thread(target=_emit_output, args=(proc.stderr,))
            t.daemon = True
            t.start()

            try:
                yield proc
            finally:
                proc.wait()
                if proc.returncode:
                    raise DataSourceError(
                        "Failed running Singer data source. Exit code: %d." % proc.returncode
                    )

    def _sync(
        self, schema: str, state: Optional[SyncState] = None, tables: Optional[TableInfo] = None
    ) -> SyncState:
        # We override the main sync() instead
        pass

    def load(self, repository: "Repository", tables: Optional[TableInfo] = None) -> str:
        return self.sync(repository, image_hash=None, tables=tables, use_state=False)

    def sync(
        self,
        repository: Repository,
        image_hash: Optional[str] = None,
        tables: Optional[TableInfo] = None,
        use_state: bool = True,
    ) -> str:
        self._validate_table_params(tables)
        tables = tables or self.tables
        config = self.get_singer_config()
        catalog = self._run_singer_discovery(config)
        catalog = self.build_singer_catalog(catalog, tables)

        base_image, new_image_hash = prepare_new_image(repository, image_hash)
        state = get_ingestion_state(repository, image_hash) if use_state else None
        logging.info("Current ingestion state: %s", state)

        # Run the sink + target and capture the stdout (new state)
        output_stream = StringIO()

        failure: Optional[Exception] = None
        try:
            with self._run_singer(config, state, catalog=catalog) as proc:
                run_patched_sync(
                    repository,
                    base_image,
                    new_image_hash,
                    delete_old=True,
                    failure="keep_both",
                    input_stream=proc.stdout,
                    output_stream=output_stream,
                )
        except DataSourceError as e:
            logging.warning("Data source partially failed. Keeping the image anyway", exc_info=e)
            failure = e

        states = output_stream.getvalue()
        latest_state = states.splitlines()[-1]
        logging.info("State stream: %s", states)

        store_ingestion_state(repository, new_image_hash, state, latest_state)

        add_timestamp_tags(repository, new_image_hash)

        repository.commit_engines()

        if failure:
            raise failure

        return new_image_hash

    def build_singer_catalog(
        self, catalog: SingerCatalog, tables: Optional[TableInfo] = None
    ) -> SingerCatalog:
        return select_streams(
            catalog, tables=tables, use_legacy_stream_selection=self.use_legacy_stream_selection
        )

    def introspect(self) -> IntrospectionResult:
        config = self.get_singer_config()
        singer_schema = self._run_singer_discovery(config)

        result = IntrospectionResult({})
        for stream in singer_schema["streams"]:
            stream_name = get_table_name(stream)
            stream_schema = get_sg_schema(stream)
            result[stream_name] = (stream_schema, cast(TableParams, {}))
        return result


def select_streams(
    catalog: SingerCatalog, tables: Optional[TableInfo] = None, use_legacy_stream_selection=False
) -> SingerCatalog:
    tables = list(tables) if tables else None

    for stream in catalog["streams"]:
        stream_name = get_table_name(stream)
        if not tables or stream_name in tables:
            if use_legacy_stream_selection:
                # See https://github.com/singer-io/getting-started/blob/5182006a2bbe542d4e94e53ddc18b59c86fcd8a2/docs/SYNC_MODE.md#legacy-streamfield-selection
                stream["schema"]["selected"] = True
            else:
                stream["metadata"][0]["metadata"]["selected"] = True

    return catalog


class GenericSingerDataSource(SingerDataSource):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.tap_path = self.params.pop("tap_path")

    @classmethod
    def get_name(cls) -> str:
        return "Generic Singer tap"

    @classmethod
    def get_description(cls) -> str:
        return "Generic Singer tap"

    def get_singer_executable(self):
        return self.tap_path

    credentials_schema = {"type": "object"}
    params_schema = {
        "type": "object",
        "properties": {"tap_path": {"type": "string"}},
        "required": ["tap_path"],
    }


class MySQLSingerDataSource(SingerDataSource):
    params_schema = {
        "type": "object",
        "properties": {
            "replication_method": {
                "type": "string",
                "enum": ["INCREMENTAL", "LOG_BASED", "FULL TABLE"],
            },
            "host": {"type": "string"},
            "port": {"type": "integer"},
        },
        "required": ["host", "port", "replication_method"],
    }

    credentials_schema = {
        "type": "object",
        "properties": {"user": {"type": "string"}, "password": {"type": "string"}},
        "required": ["user", "password"],
    }

    use_properties = True
    use_legacy_stream_selection = False

    def get_singer_executable(self):
        return "tap-mysql"

    @classmethod
    def get_name(cls) -> str:
        return "Singer MySQL tap"

    @classmethod
    def get_description(cls) -> str:
        return "Singer MySQL tap"

    def build_singer_catalog(self, catalog: SingerCatalog, tables: Optional[TableInfo] = None):
        for stream in catalog["streams"]:
            stream_name = get_table_name(stream)

            # tap-mysql requires the table metadata to contain the replication type
            if not tables or stream_name in tables:
                select_breadcrumb(stream, [])["selected"] = True

                replication_method = self.params["replication_method"]
                replication_key: Optional[str] = None

                if replication_method == "INCREMENTAL":
                    key_properties = get_key_properties(stream)
                    if not key_properties:
                        logging.warning(
                            "Table %s has replication_method INCREMENTAL but no primary key. "
                            "Falling back to FULL TABLE",
                            stream_name,
                        )
                        replication_method = "FULL TABLE"
                        replication_key = None
                    elif len(key_properties) > 1:
                        logging.warning(
                            "Table %s has a composite primary key. Falling back to FULL TABLE",
                            stream_name,
                        )
                        replication_method = "FULL TABLE"
                        replication_key = None
                    else:
                        replication_key = key_properties[0]

                select_breadcrumb(stream, [])["replication-method"] = replication_method
                if replication_key:
                    select_breadcrumb(stream, [])["replication-key"] = replication_key
        return catalog



"""
Command line routines generating example data / Splitfiles
"""
from hashlib import sha256
from typing import TYPE_CHECKING

import click

from splitgraph.commandline.common import RepositoryType
from splitgraph.core.types import TableColumn

if TYPE_CHECKING:
    from splitgraph.core.repository import Repository

_DEMO_TABLE_SIZE = 10
_DEMO_CHANGE_SIZE = 2
_DEMO_TEMPLATE = """# Import the table from the latest image in the first repository
FROM {repo_1} IMPORT {table_1} AS table_1

# Import the table from a certain image (passed in as a parameter) in the second repository
FROM {repo_2}:${{IMAGE_2}} IMPORT {table_2} AS table_2

# Create a join table
SQL CREATE TABLE result AS SELECT table_1.key, table_1.value AS value_1,\\
                                  table_2.value AS value_2\\
                           FROM table_1 JOIN table_2\\
                           ON table_1.key = table_2.key 
"""


@click.group(name="example")
def example():
    """Generate demo Splitgraph data."""


def _hash(val: int) -> str:
    """Use deterministic values to showcase reusing fragments."""
    return sha256(str(val).encode("ascii")).hexdigest()


def generate_table(repository: "Repository", table_name: str, size: int) -> None:
    """
    Creates a table with an integer primary key and a string value.

    :param repository: Checked-out Repository to create the table in.
    :param table_name: Name of the table to generate
    :param size: Number of rows in the table.
    """
    from psycopg2.sql import SQL, Identifier

    repository.engine.create_table(
        repository.to_schema(),
        table_name,
        [
            TableColumn(1, "key", "integer", True, "Some key"),
            TableColumn(2, "value", "varchar", False, "Some value"),
        ],
    )
    repository.engine.run_sql_batch(
        SQL("INSERT INTO {} VALUES (%s, %s)").format(Identifier(table_name)),
        [(i, _hash(i)) for i in range(size)],
        schema=repository.to_schema(),
    )


def alter_table(
    repository: "Repository", table_name: str, rows_added: int, rows_deleted: int, rows_updated: int
) -> None:
    """
    Alters the example table, adding/updating/deleting a certain number of rows.

    :param repository: Checked-out Repository object.
    :param table_name: Name of the table
    :param rows_added: Number of rows to add
    :param rows_deleted: Number of rows to remove
    :param rows_updated: Number of rows to update
    """
    from psycopg2.sql import SQL, Identifier

    from splitgraph.core.sql import select
    from splitgraph.engine import ResultShape

    keys = repository.run_sql(
        select(table_name, "key", schema=repository.to_schema()), return_shape=ResultShape.MANY_ONE
    )
    last = repository.run_sql(
        select(table_name, "MAX(key)", schema=repository.to_schema()),
        return_shape=ResultShape.ONE_ONE,
    )

    # Delete first N rows
    click.echo("Deleting %d rows..." % rows_deleted)
    repository.engine.run_sql_batch(
        SQL("DELETE FROM {} WHERE key = %s").format(Identifier(table_name)),
        [(k,) for k in keys[:rows_deleted]],
        schema=repository.to_schema(),
    )

    # Update next N rows
    click.echo("Updating %d rows..." % rows_updated)
    repository.engine.run_sql_batch(
        SQL("UPDATE {} SET value = %s WHERE key = %s").format(Identifier(table_name)),
        [(_hash(k) + "_UPDATED", k) for k in keys[rows_updated : rows_updated * 2]],
        schema=repository.to_schema(),
    )

    # Insert rows at the end
    click.echo("Adding %d rows..." % rows_added)
    repository.engine.run_sql_batch(
        SQL("INSERT INTO {} VALUES (%s, %s)").format(Identifier(table_name)),
        [(k, _hash(k)) for k in range(last + 1, last + rows_added + 1)],
        schema=repository.to_schema(),
    )


@click.command(name="generate")
@click.argument("repository", type=RepositoryType())
def generate_c(repository):
    """
    Generate a repository with some example data.

    :param repository: Repository to generate. Must not already exist.
    """
    from splitgraph.core.engine import repository_exists

    if repository_exists(repository):
        raise click.ClickException(
            "Repository %s already exists, use sgr rm to delete it!" % repository.to_schema()
        )

    repository.init()
    # Insert some data
    generate_table(repository, "demo", size=_DEMO_TABLE_SIZE)

    image = repository.commit()
    click.echo(
        "Generated %s:%s with %s rows, image hash %s."
        % (repository.to_schema(), "demo", _DEMO_TABLE_SIZE, image.image_hash[:12])
    )


@click.command(name="alter")
@click.argument("repository", type=RepositoryType(exists=True))
def alter_c(repository):
    """
    Alter the table in an example repository.

    This deletes the first 2 rows, updates the next 2 rows and adds 2 rows at the end. Doesn't create a new image.

    :param repository: Generated demo repository.
    """
    alter_table(repository, "demo", _DEMO_CHANGE_SIZE, _DEMO_CHANGE_SIZE, _DEMO_CHANGE_SIZE)


@click.command(name="splitfile")
@click.argument("repository_1", type=RepositoryType())
@click.argument("repository_2", type=RepositoryType())
def splitfile_c(repository_1, repository_2):
    """
    Generate a sample Splitfile.

    The Splitfile imports tables from two example repositories (generated by sgr generate/alter) and runs a
    JOIN query on them.

    :param repository_1: First repository
    :param repository_2: Second repository
    """
    click.echo(
        _DEMO_TEMPLATE.format(
            repo_1=repository_1.to_schema(),
            repo_2=repository_2.to_schema(),
            table_1="demo",
            table_2="demo",
        )
    )


example.add_command(generate_c)
example.add_command(alter_c)
example.add_command(splitfile_c)



from abc import ABCMeta, abstractmethod
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    List,
    NamedTuple,
    NewType,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
)

from pydantic import BaseModel

if TYPE_CHECKING:
    from splitgraph.cloud.models import ExternalTableRequest

Changeset = Dict[Tuple[str, ...], Tuple[bool, Dict[str, Any], Dict[str, Any]]]


class TableColumn(NamedTuple):
    ordinal: int
    name: str
    pg_type: str
    is_pk: bool
    comment: Optional[str] = None


TableSchema = List[TableColumn]
Quals = Sequence[Sequence[Tuple[str, str, Any]]]

SourcesList = List[Dict[str, str]]
ProvenanceLine = Dict[str, Union[str, List[str], List[bool], SourcesList]]

# Ingestion-related params
Credentials = NewType("Credentials", Dict[str, Any])
Params = NewType("Params", Dict[str, Any])
TableParams = NewType("TableParams", Dict[str, Any])
TableInfo = Union[List[str], Dict[str, Tuple[TableSchema, TableParams]]]
SyncState = NewType("SyncState", Dict[str, Any])


class MountError(BaseModel):
    table_name: str
    error: str
    error_text: str


PreviewResult = NewType("PreviewResult", Dict[str, Union[MountError, List[Dict[str, Any]]]])
IntrospectionResult = NewType(
    "IntrospectionResult", Dict[str, Union[Tuple[TableSchema, TableParams], MountError]]
)

T = TypeVar("T")


def unwrap(
    result: Dict[str, Union[MountError, T]],
) -> Tuple[Dict[str, T], Dict[str, MountError]]:
    good = {}
    bad = {}
    for k, v in result.items():
        if isinstance(v, MountError):
            bad[k] = v
        else:
            good[k] = v
    return good, bad


def get_table_params(table_info: TableInfo, table_name: str) -> TableParams:
    if isinstance(table_info, dict) and table_name in table_info:
        return table_info[table_name][1]
    return TableParams({})


def get_table_list(table_info: TableInfo) -> List[str]:
    if isinstance(table_info, list):
        return table_info
    return list(table_info.keys())


class Comparable(metaclass=ABCMeta):
    @abstractmethod
    def __lt__(self, other: Any) -> bool:
        ...


def dict_to_table_schema_params(
    tables: Dict[str, "ExternalTableRequest"]
) -> Dict[str, Tuple[TableSchema, TableParams]]:
    return {
        t: (
            [
                TableColumn(i + 1, cname, ctype, False, None)
                for (i, (cname, ctype)) in enumerate((tsp.schema_ or {}).items())
            ],
            TableParams(tsp.options),
        )
        for t, tsp in tables.items()
    }


def table_schema_params_to_dict(
    tables: Dict[str, Tuple[TableSchema, TableParams]]
) -> Dict[str, Dict[str, Dict[str, Any]]]:
    return {
        t: {
            "schema": {c.name: c.pg_type for c in ts},
            "options": tp,
        }
        for t, (ts, tp) in tables.items()
    }



from configparser import ConfigParser, ExtendedInterpolation
from typing import Dict, Union, cast

from splitgraph.config.keys import ConfigDict


def hoist_section(
    config_dict: Dict[str, Dict[str, str]], section: str = "defaults"
) -> Dict[str, Union[Dict[str, str], str]]:
    """
    If a section exists called <hoist_section>, hoist it to the top level
    This is useful for overriding default configs from within a config file

    Transform `config_dict` to "hoist" any config values from a section
    into the top level (thus, overriding environment variables),
    when the name of the section matches `hoist_section`.

    Return a new, updated copy of `config_dict`.
    """

    if section not in config_dict.keys():
        # see https://github.com/python/mypy/issues/2300
        return cast(Dict[str, Union[Dict[str, str], str]], config_dict)

    hoisted_config_dict = cast(Dict[str, Union[Dict[str, str], str]], config_dict.copy())
    section_dict = config_dict[section].copy()
    del hoisted_config_dict[section]

    for k, v in section_dict.items():
        hoisted_config_dict[k.upper()] = v

    return hoisted_config_dict


def accumulate_lists(config_dict: Dict[str, Union[Dict[str, str], str]]) -> ConfigDict:
    """
    Transform a `config_dict` to "accumulate" objects "nested" via key name

    Because  ConfigParser does not support nesting, we implement our own
    syntax via the key names of sections. The ':' character can be used in
    section names to specify the left and right key. Example::

        .ini config             new config_dict

                          ----->
                            |
        [remote: remote1]   |   {
        SG_ENV_VAR=foo      |       **rest_of_config_dict,
                            |       "remotes": {
        [origin: origin1]   |           "remote1": {
        SG_ENV_VAR=bar      |               "SG_ENV_VAR": "foo"
                            |           }
        [origin: origin2]   |       },
        SG_ENV_VAR=bax      |       "origins": {
                            |           "origin1": {
                            |               "SG_ENV_VAR": "bar"
                            |           },
                            |           "origin2": {
                            |               "SG_ENV_VAR": "bax"
                            |           }
                            |       }
                            |   }
                            |
                          ----->

    :return a new, updated copy of `config_dict`
    """

    new_dict = cast(ConfigDict, config_dict.copy())

    accumulatable = {"remote": "remotes", "origin": "origins"}

    accumulatable_keys = accumulatable.keys()

    accumulated: Dict[str, Dict[str, Dict[str, str]]] = {}

    def key_matches(k):
        return k.split(":")[0] in accumulatable_keys

    matching_keys = [k for k in new_dict.keys() if key_matches(k)]

    keys_to_delete = []

    # e.g. key = "remote: remote1"
    for key in matching_keys:
        keys_to_delete.append(key)

        # e.g. "remote"
        left_key = key.split(":")[0]

        # e.g. "remote1" (to be used as a key in the new dict)
        right_key = "".join(key.split(":")[1:]).strip()

        list_key = accumulatable[left_key]
        new_item = cast(Dict[str, str], new_dict[key]).copy()
        if list_key not in accumulated.keys():
            accumulated[list_key] = {right_key: new_item}
        else:
            accumulated[list_key][right_key] = new_item

    for delete_key in keys_to_delete:
        del new_dict[delete_key]

    new_dict.update(accumulated)

    return new_dict


def transform_config_dict(config_dict: Dict[str, Dict[str, str]], **kwargs) -> ConfigDict:
    """
    Apply transformations to the raw ConfigParser.config object

        1) hoist_section
        2) accumulate_lists

    Return the a new, updated copy of `config_dict`.
    """

    config_dict = hoist_section(config_dict, **kwargs)
    config_dict = accumulate_lists(config_dict)
    return config_dict


# TODO: Get this cleaned up for parsing environment variables as default
#
# class EnvInterpolation(ExtendedInterpolation):
#     def before_get(self, parser, section, option, value, defaults):
#         return os.path.expandvars(value)
#         # if expanded == value:
#         #     return super(ExtendedInterpolation, self).before_get(parser, section, option, value, defaults)
#         # else:
#         #     return expanded


def get_config_dict_from_file(sg_file: str, **kwargs) -> Dict[str, Dict[str, str]]:
    # TODO
    # config = ConfigParser(interpolation=EnvInterpolation())
    config = ConfigParser(interpolation=ExtendedInterpolation())

    # Convert all keys to uppercase
    config.optionxform = lambda option: option.upper()  # type: ignore
    config.read(sg_file)

    config_dict = {s: dict(config.items(s, False)) for s in config.sections()}

    # Fix certain sections to have lower case by convention
    if "data_sources" in config_dict:
        handlers = config_dict["data_sources"]
        assert isinstance(handlers, dict)
        config_dict["data_sources"] = cast(
            Dict[str, str], {k.lower(): v for k, v in handlers.items()}
        )

    return config_dict


def get_config_dict_from_config_file(sg_file: str, **kwargs) -> ConfigDict:
    """
    Create a dict from ConfigParser, apply transformations to it.

    Return parsed and transformed `config_dict`.
    """

    config_dict = get_config_dict_from_file(sg_file, **kwargs)
    config_dict = transform_config_dict(config_dict, **kwargs)

    return config_dict



"""
Common internal functions used by Splitgraph commands.
"""
import importlib.resources
import logging
from datetime import date, datetime, time
from decimal import Decimal
from functools import wraps
from random import getrandbits
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Set,
    Tuple,
    Union,
    cast,
)

from psycopg2.sql import SQL, Identifier

from splitgraph.config import SPLITGRAPH_API_SCHEMA, SPLITGRAPH_META_SCHEMA
from splitgraph.core.migration import set_installed_version, source_files_to_apply
from splitgraph.core.output import parse_date, parse_dt
from splitgraph.core.sql import select
from splitgraph.exceptions import ImageNotFoundError, RepositoryNotFoundError
from splitgraph.resources import splitgraph_meta

if TYPE_CHECKING:
    from splitgraph.core.repository import Repository
    from splitgraph.engine.postgres.engine import PostgresEngine, PsycopgEngine

META_TABLES = [
    "images",
    "tags",
    "objects",
    "tables",
    "upstream",
    "object_locations",
    "object_cache_status",
    "object_cache_occupancy",
    "info",
    "version",
]
OBJECT_MANAGER_TABLES = ["object_cache_status", "object_cache_occupancy"]


def set_tag(repository: "Repository", image_hash: Optional[str], tag: str) -> None:
    """Internal function -- add a tag to an image."""
    engine = repository.engine
    engine.run_sql(
        SQL("SELECT {}.tag_image(%s,%s,%s,%s)").format(Identifier(SPLITGRAPH_API_SCHEMA)),
        (repository.namespace, repository.repository, image_hash, tag),
    )


def set_tags_batch(repository: "Repository", hashes_tags: List[Tuple[str, str]]) -> None:
    engine = repository.engine

    args = [
        (repository.namespace, repository.repository, image_hash, tag)
        for image_hash, tag in hashes_tags
    ]

    engine.run_sql_batch(
        SQL("SELECT {}.tag_image(%s,%s,%s,%s)").format(Identifier(SPLITGRAPH_API_SCHEMA)),
        args,
    )


def set_head(repository: "Repository", image: Optional[str]) -> None:
    """Sets the HEAD pointer of a given repository to a given image. Shouldn't be used directly."""
    set_tag(repository, image, "HEAD")


def manage_audit_triggers(
    engine: "PostgresEngine", object_engine: Optional["PostgresEngine"] = None
) -> None:
    """Does bookkeeping on audit triggers / audit table:

        * Detect tables that are being audited that don't need to be any more
          (e.g. they've been unmounted)
        * Drop audit triggers for those and delete all audit info for them
        * Set up audit triggers for new tables

    If the metadata engine isn't the same as the object engine, this does nothing.

    :param engine: Metadata engine with information about images and their checkout state
    :param object_engine: Object engine where the checked-out table and the audit triggers are located.
    """

    object_engine = object_engine or engine

    if object_engine != engine:
        return

    from splitgraph.core.engine import get_current_repositories

    repos_tables = [
        (r.to_schema(), t)
        for r, head in get_current_repositories(engine)
        if head is not None
        for t in set(object_engine.get_all_tables(r.to_schema())) & set(head.get_tables())
    ]
    tracked_tables = object_engine.get_tracked_tables()

    to_untrack = [t for t in tracked_tables if t not in repos_tables]
    to_track = [t for t in repos_tables if t not in tracked_tables]

    if to_untrack:
        object_engine.untrack_tables(to_untrack)

    if to_track:
        object_engine.track_tables(to_track)


def manage_audit(func: Callable) -> Callable:
    """A decorator to be put around various Splitgraph commands
    that adds/removes audit triggers for new/committed/deleted tables.
    """

    # Make sure docstrings are passed through
    @wraps(func)
    def wrapped(self, *args, **kwargs):
        from .image import Image

        if isinstance(self, Image):
            repository = self.repository
        else:
            repository = self
        try:
            manage_audit_triggers(repository.engine, repository.object_engine)
            return func(self, *args, **kwargs)
        finally:
            repository.object_engine.commit()
            manage_audit_triggers(repository.engine, repository.object_engine)

    return wrapped


def ensure_metadata_schema(engine: "PsycopgEngine") -> None:
    """
    Create or migrate the metadata schema splitgraph_meta that stores the hash tree of schema
    snapshots (images), tags and tables.
    This means we can't mount anything under the schema splitgraph_meta -- much like we can't have a folder
    ".git" under Git version control...
    """

    schema_files = [f for f in importlib.resources.contents(splitgraph_meta) if f.endswith("sql")]

    files, target_version = source_files_to_apply(
        engine,
        schema_name=SPLITGRAPH_META_SCHEMA,
        schema_files=schema_files,
    )

    if not files:
        return
    for name in files:
        data = importlib.resources.read_text(splitgraph_meta, name)
        logging.info("Running %s", name)
        engine.run_sql(data)
    set_installed_version(engine, SPLITGRAPH_META_SCHEMA, target_version)
    engine.commit()


def aggregate_changes(
    query_result: List[Tuple[int, int]], initial: Optional[Tuple[int, int, int]] = None
) -> Tuple[int, int, int]:
    """Add a changeset to the aggregated diff result"""
    result = list(initial) if initial else [0, 0, 0]
    for kind, kind_count in query_result:
        assert kind in (0, 1, 2)
        result[kind] += kind_count
    return result[0], result[1], result[2]


def slow_diff(
    repository: "Repository",
    table_name: str,
    image_1: Optional[str],
    image_2: Optional[str],
    aggregate: bool,
) -> Union[Tuple[int, int, int], List[Tuple[bool, Tuple]]]:
    """Materialize both tables and manually diff them"""
    with repository.materialized_table(table_name, image_1) as (mp_1, table_1):
        with repository.materialized_table(table_name, image_2) as (mp_2, table_2):
            # Check both tables out at the same time since then table_2 calculation can be based
            # on table_1's snapshot.
            left = repository.object_engine.run_sql(
                SQL("SELECT * FROM {}.{}").format(Identifier(mp_1), Identifier(table_1))
            )
            right = repository.object_engine.run_sql(
                SQL("SELECT * FROM {}.{}").format(Identifier(mp_2), Identifier(table_2))
            )

    if aggregate:
        return sum(1 for r in right if r not in left), sum(1 for r in left if r not in right), 0

    # Return format: list of [(False for deleted/True for inserted, full row)]
    return [(False, r) for r in left if r not in right] + [
        (True, r) for r in right if r not in left
    ]


def gather_sync_metadata(
    target: "Repository",
    source: "Repository",
    overwrite_objects=False,
    overwrite_tags=False,
    single_image: Optional[str] = None,
) -> Any:
    """
    Inspects two Splitgraph repositories and gathers metadata that is required to bring target up to
    date with source.

    :param target: Target Repository object
    :param source: Source repository object
    :param overwrite_objects: If True, will return metadata for all objects
        belonging to new images (or existing image if single_image=True)
    :param single_image: If set, only grab a single image with this hash/tag from the source.
    :param overwrite_tags: If True and single_image is set, will return all tags for that image.
        If single_image is not set, will return all tags in the source repository.
        If False, will only return tags in the source that don't exist on the target.

    :returns: Tuple of metadata for  new_images, new_tables, object_locations, object_meta, tags
    """

    # Transaction handling: this is run as a single read-only transaction to the source and the
    # target engine. This is so that:
    #   1) we can in the future route these to read-only replicas
    #   2) we don't hold an open transaction to the registry while we're
    #       downloading/uploading objects.
    target.engine.run_sql("SET TRANSACTION READ ONLY")
    if target.engine != source.engine:
        source.engine.run_sql("SET TRANSACTION READ ONLY")

    try:
        result = _gather_sync_metadata(
            target, source, overwrite_objects, overwrite_tags, single_image
        )
        target.commit_engines()
        source.commit_engines()
        return result
    except Exception:
        target.rollback_engines()
        source.rollback_engines()
        raise


def _gather_sync_metadata(target, source, overwrite_objects, overwrite_tags, single_image) -> Any:
    # Currently, images can't be altered once pushed out. We intend to relax this:
    # same image hash means same contents and same tables but the composition of an image
    # can change (if we refragment a table so that querying it is faster). But it's frowned
    # upon.
    single_image_hash: Optional[str] = None
    if single_image:
        image = source.images[single_image]
        single_image_hash = image.image_hash
        try:
            # If the image already exists on the target, we shouldn't overwrite it.
            # The user can get around this by deleting the image manually.
            _ = target.images[single_image_hash]
            new_images = []
            new_image_hashes = []
        except ImageNotFoundError:
            new_images = [image]
            new_image_hashes = [single_image_hash]
        except RepositoryNotFoundError:
            new_images = [image]
            new_image_hashes = [single_image_hash]
    else:
        # If an image hasn't been specified, get/push all non-existing images.
        target_images = {i.image_hash: i for i in target.images()}
        source_images = {i.image_hash: i for i in source.images()}
        new_image_hashes = [i for i in source_images if i not in target_images]
        new_images = [source_images[i] for i in new_image_hashes]

    # Get the meta for all tables we'll need to fetch.
    table_meta = []

    # Also grab the list of all objects in this repository in case overwrite_objects=True
    all_objects: Set[str] = set()
    for t in source.engine.run_sql(
        select(
            "get_all_tables",
            "image_hash, table_name, table_schema, object_ids",
            schema=SPLITGRAPH_API_SCHEMA,
            table_args="(%s,%s)",
        ),
        (source.namespace, source.repository),
    ):
        if t[0] in new_image_hashes:
            all_objects = all_objects.union(t[-1])
            table_meta.append(t)
        elif t[0] == single_image_hash:
            # Overwrite objects from existing image if it was
            # passed as single_image (similar behaviour to tags).
            all_objects = all_objects.union(t[-1])

    # Get the tags
    existing_tags = [t for s, t in target.get_all_hashes_tags()]
    tags = {
        t: s
        for s, t in source.get_all_hashes_tags()
        if (
            # Only get new tags (unless we're overwriting them)
            t not in existing_tags
            or overwrite_tags
        )
        and (
            # Only get tags for the new image (unless we're pulling the whole repo)
            not single_image
            or s == single_image_hash
        )
    }

    # Get objects that don't exist on the target
    table_objects = list({o for table in table_meta for o in table[3]})
    new_objects = list(set(target.objects.get_new_objects(table_objects)))

    if overwrite_objects:
        new_objects = list(all_objects)

    object_locations = source.objects.get_external_object_locations(new_objects)

    if new_objects:
        object_meta = source.objects.get_object_meta(new_objects)
    else:
        object_meta = {}
    return new_images, table_meta, object_locations, object_meta, tags


_TYPE_MAP: Dict[str, Callable] = {
    k: cast(Callable, v)
    for ks, v in [
        (["integer", "bigint", "smallint"], int),
        (["numeric", "real", "double precision"], float),
        (["timestamp", "timestamp without time zone"], parse_dt),
        (["date"], parse_date),
    ]
    for k in ks
}


def adapt(value: Any, pg_type: str) -> Any:
    """
    Coerces a value with a PG type into its Python equivalent.

    :param value: Value
    :param pg_type: Postgres datatype
    :return: Coerced value.
    """
    if value is None:
        return None
    if not isinstance(value, str):
        return value
    if pg_type in _TYPE_MAP:
        return _TYPE_MAP[pg_type](value)
    return value


class Tracer:
    """
    Accumulates events and returns the times between them.
    """

    def __init__(self) -> None:
        self.start_time = datetime.now()
        self.events: List[Tuple[datetime, str]] = []

    def log(self, event: str) -> None:
        """
        Log an event at the current time
        :param event: Event name
        """
        self.events.append((datetime.now(), event))

    def get_durations(self) -> List[Tuple[str, float]]:
        """
        Return all events and durations between them.
        :return: List of (event name, time to this event from the previous event (or start))
        """
        result = []
        prev = self.start_time
        for event_time, event in self.events:
            result.append((event, (event_time - prev).total_seconds()))
            prev = event_time
        return result

    def get_total_time(self) -> float:
        """
        :return: Time from start to the final logged event.
        """
        return (self.events[-1][0] - self.start_time).total_seconds()

    def __str__(self) -> str:
        result = ""
        for event, duration in self.get_durations():
            result += "\n%s: %.3f" % (event, duration)
        result += "\nTotal: %.3f" % self.get_total_time()
        return result[1:]


def coerce_val_to_json(val: Any) -> Any:
    """
    Turn a Python value to a string/float that can be stored as JSON.
    """
    if isinstance(val, list):
        val = [coerce_val_to_json(v) for v in val]
    elif isinstance(val, tuple):
        val = tuple(coerce_val_to_json(v) for v in val)
    elif isinstance(val, dict):
        val = {k: coerce_val_to_json(v) for k, v in val.items()}
    elif isinstance(val, (Decimal, date, time, datetime)):
        # See https://www.postgresql.org/docs/11/datatype-datetime.html
        # "ISO 8601 specifies the use of uppercase letter T to separate the date and time.
        # PostgreSQL accepts that format on input, but on output it uses a space rather
        # than T, as shown above. "
        #
        # This also matches python's str().
        return str(val)
    return val


class CallbackList(list):
    """
    Used to pass around and call multiple callbacks at once.
    """

    def __call__(self, *args, **kwargs) -> None:
        for listener in self:
            listener(*args, **kwargs)


def get_temporary_table_id() -> str:
    """Generate a random ID for temporary/staging objects that haven't had their ID calculated yet."""
    return str.format("sg_tmp_{:032x}", getrandbits(128))


def unmount_schema(engine: "PsycopgEngine", schema: str) -> None:
    engine.run_sql(
        SQL("DROP SERVER IF EXISTS {} CASCADE").format(Identifier("%s_lq_checkout_server" % schema))
    )
    engine.run_sql(SQL("DROP SERVER IF EXISTS {} CASCADE").format(Identifier(schema + "_server")))
    engine.run_sql(SQL("DROP SCHEMA IF EXISTS {} CASCADE").format(Identifier(schema)))



"""
Classes related to managing table/image/object metadata tables.
"""
import itertools
from datetime import datetime
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    List,
    NamedTuple,
    Optional,
    Sequence,
    Tuple,
    cast,
)

from psycopg2.extras import Json
from psycopg2.sql import SQL, Identifier

from splitgraph.config import SPLITGRAPH_API_SCHEMA, SPLITGRAPH_META_SCHEMA
from splitgraph.core.types import TableSchema
from splitgraph.engine import ResultShape
from splitgraph.engine.postgres.engine import API_MAX_VARIADIC_ARGS, chunk

from .sql import select

if TYPE_CHECKING:
    from splitgraph.core.repository import Repository
    from splitgraph.engine.postgres.engine import PsycopgEngine

OBJECT_COLS = [
    "object_id",
    "format",
    "namespace",
    "size",
    "created",
    "insertion_hash",
    "deletion_hash",
    "index",
    "rows_inserted",
    "rows_deleted",
]


class Object(NamedTuple):
    """Represents a Splitgraph object that tables are composed of."""

    object_id: str
    format: str
    namespace: str
    size: int
    created: datetime
    insertion_hash: str
    deletion_hash: str
    object_index: Dict[str, Any]  # Clashes with NamedTuple's "index"
    rows_inserted: int
    rows_deleted: int


class MetadataManager:
    """
    A data access layer for the metadata tables in the splitgraph_meta schema that concerns itself
    with image, table and object information.
    """

    def __init__(self, metadata_engine: "PsycopgEngine") -> None:
        self.metadata_engine = metadata_engine

    def register_objects(self, objects: List[Object], namespace: Optional[str] = None) -> None:
        """
        Registers multiple Splitgraph objects in the tree.

        :param objects: List of `Object` objects.
        :param namespace: If specified, overrides the original object namespace, required
            in the case where the remote repository has a different namespace than the local one.
        """
        object_meta = [
            tuple(
                namespace
                if namespace and a == "namespace"
                else getattr(o, a if a != "index" else "object_index")
                for a in OBJECT_COLS
            )
            for o in sorted(objects, key=lambda o: o.object_id)
        ]

        self.metadata_engine.run_sql_batch(
            SQL(
                "SELECT {}.add_object(" + ",".join(itertools.repeat("%s", len(OBJECT_COLS))) + ")"
            ).format(Identifier(SPLITGRAPH_API_SCHEMA)),
            object_meta,
        )

    def register_tables(
        self, repository: "Repository", table_meta: List[Tuple[str, str, TableSchema, List[str]]]
    ) -> None:
        """
        Links tables in an image to physical objects that they are stored as.
        Objects must already be registered in the object tree.

        :param repository: Repository that the tables belong to.
        :param table_meta: A list of (image_hash, table_name, table_schema, object_ids).
        """
        table_meta = [
            (repository.namespace, repository.repository, o[0], o[1], Json(o[2]), o[3])
            for o in table_meta
        ]

        # We're about to send len(table_meta) statements to the API and if one of them
        # is larger than the max query length, this won't work. We can't use run_chunk_sql
        # either since that only works on one statement and we might have several here
        # (still want to minimize the number of roundtrips). So we chunk the parameters
        # ourselves.

        rechunked_meta = []
        for meta in table_meta:
            rest, objects = meta[:-1], meta[-1]
            if len(objects) <= API_MAX_VARIADIC_ARGS:
                rechunked_meta.append(meta)
                continue
            rechunked_meta.extend(
                [rest + (c,) for c in chunk(objects, chunk_size=API_MAX_VARIADIC_ARGS)]
            )

        self.metadata_engine.run_sql_batch(
            SQL("SELECT {}.add_table(%s,%s,%s,%s,%s,%s)").format(Identifier(SPLITGRAPH_API_SCHEMA)),
            rechunked_meta,
        )

    def overwrite_table(
        self,
        repository: "Repository",
        image_hash: str,
        table_name: str,
        table_schema: TableSchema,
        objects: List[str],
    ):
        self.metadata_engine.run_sql(
            "UPDATE splitgraph_meta.tables "
            "SET table_schema = %s, object_ids = %s "
            "WHERE namespace = %s AND repository = %s AND image_hash = %s "
            "AND table_name = %s",
            (
                Json(table_schema),
                objects,
                repository.namespace,
                repository.repository,
                image_hash,
                table_name,
            ),
        )

    def register_object_locations(self, object_locations: List[Tuple[str, str, str]]) -> None:
        """
        Registers external locations (e.g. HTTP or S3) for Splitgraph objects.
        Objects must already be registered in the object tree.

        :param object_locations: List of (object_id, location, protocol).
        """
        self.metadata_engine.run_sql_batch(
            SQL("SELECT {}.add_object_location(%s,%s,%s)").format(
                Identifier(SPLITGRAPH_API_SCHEMA)
            ),
            sorted(object_locations, key=lambda o: str(o[0])),
        )

    def get_all_objects(self) -> List[str]:
        """
        Gets all objects currently in the Splitgraph tree.

        :return: List of object IDs.
        """
        return cast(
            List[str],
            self.metadata_engine.run_sql(
                select("objects", "object_id"), return_shape=ResultShape.MANY_ONE
            ),
        )

    def get_new_objects(self, object_ids: List[str]) -> List[str]:
        """
        Get object IDs from the passed list that don't exist in the tree.

        :param object_ids: List of objects to check
        :return: List of unknown object IDs.
        """
        return cast(
            List[str],
            self.metadata_engine.run_chunked_sql(
                SQL("SELECT {}.get_new_objects(%s)").format(Identifier(SPLITGRAPH_API_SCHEMA)),
                (object_ids,),
                return_shape=ResultShape.ONE_ONE,
                chunk_position=0,
            ),
        )

    def get_external_object_locations(self, objects: List[str]) -> List[Tuple[str, str, str]]:
        """
        Gets external locations for objects.

        :param objects: List of object IDs stored externally.
        :return: List of (object_id, location, protocol).
        """
        return cast(
            List[Tuple[str, str, str]],
            self.metadata_engine.run_chunked_sql(
                select(
                    "get_object_locations",
                    "object_id, location, protocol",
                    schema=SPLITGRAPH_API_SCHEMA,
                    table_args="(%s)",
                ),
                (objects,),
                chunk_position=0,
            ),
        )

    def get_object_meta(self, objects: List[str]) -> Dict[str, Object]:
        """
        Get metadata for multiple Splitgraph objects from the tree

        :param objects: List of objects to get metadata for.
        :return: Dictionary of object_id -> Object
        """
        if not objects:
            return {}

        metadata = self.metadata_engine.run_chunked_sql(
            select(
                "get_object_meta",
                ",".join(OBJECT_COLS),
                schema=SPLITGRAPH_API_SCHEMA,
                table_args="(%s)",
            ),
            (list(objects),),
            chunk_position=0,
        )
        result = [Object(*m) for m in metadata]
        return {o.object_id: o for o in result}

    def get_objects_for_repository(
        self, repository: "Repository", image_hash: Optional[str] = None
    ) -> List[str]:
        query = SQL(
            "SELECT object_ids FROM {}.tables WHERE namespace = %s AND repository = %s"
        ).format(Identifier(SPLITGRAPH_META_SCHEMA))
        args = [repository.namespace, repository.repository]
        if image_hash:
            query += SQL(" AND image_hash = %s")
            args.append(image_hash)

        table_objects = {
            o
            for os in self.metadata_engine.run_sql(
                query,
                args,
                return_shape=ResultShape.MANY_ONE,
            )
            for o in os
        }

        return list(table_objects)

    def delete_object_meta(self, object_ids: Sequence[str]):
        """
        Delete metadata for multiple objects (external locations, indexes, hashes).
        This doesn't delete physical objects.

        :param object_ids: Object IDs to delete
        """
        for table_name in ["object_locations", "objects"]:
            self.metadata_engine.run_chunked_sql(
                SQL("DELETE FROM {}.{} WHERE object_id = ANY(%s)").format(
                    Identifier(SPLITGRAPH_META_SCHEMA), Identifier(table_name)
                ),
                arguments=(object_ids,),
                chunk_position=0,
            )

    def get_unused_objects(self, threshold: Optional[int] = None) -> List[Tuple[str, datetime]]:
        """
        Get a list of all objects in the metadata that aren't used by any table and can be
        safely deleted.

        :param threshold: Only return objects that were created earlier than this (in minutes)
        :return: List of objects and their creation times.
        """
        candidates = self.metadata_engine.run_sql(
            SQL(  # nosec
                "SELECT object_id, created FROM {0}.objects "
                "WHERE object_id NOT IN (SELECT unnest(object_ids) "
                "FROM {0}.tables) "
                + ("AND created < now() - INTERVAL '%s minutes' " if threshold is not None else "")
                + "ORDER BY created"
            ).format(Identifier(SPLITGRAPH_META_SCHEMA)),
            (threshold,) if threshold is not None else None,
        )
        return cast(List[Tuple[str, datetime]], candidates)

    def cleanup_metadata(self) -> List[str]:
        """
        Go through the current metadata and delete all objects that aren't required
        by any table on the engine.

        :return: List of objects that have been deleted.
        """
        candidates = self.get_unused_objects()
        to_delete = [c for c, _ in candidates]

        self.delete_object_meta(to_delete)
        return to_delete



"""Definitions for the splitgraph.yml format that's used to batch-populate a Splitgraph catalog
with repositories and their metadata.
"""

from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel, Field


class Table(BaseModel):
    class Column(BaseModel):
        name: str
        pg_type: str = Field(alias="type")
        comment: Optional[str]

    options: Dict[str, Any]
    schema_: List[Column] = Field(alias="schema")


class Credential(BaseModel):
    plugin: str
    data: Dict[str, Any]


class IngestionSchedule(BaseModel):
    schedule: str
    enabled = True


class External(BaseModel):
    credential_id: Optional[str]
    credential: Optional[str]
    plugin: str
    params: Dict[str, Any]
    tables: Dict[str, Table]
    is_live: bool = True
    schedule: Optional[IngestionSchedule]


class Source(BaseModel):
    anchor: str
    href: str
    isCreator: Optional[bool]
    isSameAs: Optional[bool]


class Metadata(BaseModel):
    class Readme(BaseModel):
        file: Optional[str]
        text: Optional[str]

    readme: Optional[Union[str, Readme]]
    description: Optional[str]
    topics: Optional[List[str]]
    sources: Optional[List[Source]]
    license: Optional[str]
    extra_metadata: Optional[Dict[str, Any]]


class Repository(BaseModel):
    namespace: str
    repository: str
    metadata: Optional[Metadata]
    external: Optional[External]


class SplitgraphYAML(BaseModel):
    repositories: List[Repository] = []
    credentials: Optional[Dict[str, Credential]]



import contextlib
import json
from copy import deepcopy
from datetime import timedelta
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, cast

from psycopg2.sql import SQL, Identifier

from splitgraph.core.types import (
    Credentials,
    MountError,
    Params,
    TableInfo,
    TableParams,
    TableSchema,
)
from splitgraph.hooks.data_source.fdw import (
    ForeignDataWrapperDataSource,
    import_foreign_schema,
)
from splitgraph.ingestion.common import IngestionAdapter, build_commandline_help
from splitgraph.ingestion.csv.common import dump_options, load_options

if TYPE_CHECKING:
    from splitgraph.engine.postgres.engine import PostgresEngine, PsycopgEngine


class CSVIngestionAdapter(IngestionAdapter):
    @staticmethod
    def create_ingestion_table(data, engine, schema: str, table: str, **kwargs):
        schema_spec = kwargs.pop("schema_spec")
        engine.delete_table(schema, table)
        engine.create_table(schema, table, schema_spec=schema_spec, include_comments=True)

    @staticmethod
    def data_to_new_table(
        data, engine: "PostgresEngine", schema: str, table: str, no_header: bool = True, **kwargs
    ):
        copy_csv_buffer(data, engine, schema, table, no_header, **kwargs)

    @staticmethod
    def query_to_data(engine, query: str, schema: Optional[str] = None, **kwargs):
        buffer = kwargs.pop("buffer")
        query_to_csv(engine, query, buffer, schema)
        return buffer


csv_adapter = CSVIngestionAdapter()


def copy_csv_buffer(
    data, engine: "PsycopgEngine", schema: str, table: str, no_header: bool = False, **kwargs
):
    """Copy CSV data from a buffer into a given schema/table"""
    with engine.copy_cursor() as cur:
        extra_args = [not no_header]

        copy_command = SQL("COPY {}.{} FROM STDIN WITH (FORMAT CSV, HEADER %s").format(
            Identifier(schema), Identifier(table)
        )
        for k, v in kwargs.items():
            if k in ("encoding", "delimiter", "escape") and v:
                copy_command += SQL(", " + k + " %s")
                extra_args.append(v)
        copy_command += SQL(")")

        cur.copy_expert(
            cur.mogrify(copy_command, extra_args),
            data,
        )


def query_to_csv(engine: "PsycopgEngine", query, buffer, schema: Optional[str] = None):
    copy_query = SQL("COPY (") + SQL(query) + SQL(") TO STDOUT WITH (FORMAT CSV, HEADER TRUE);")
    if schema:
        copy_query = SQL("SET search_path TO {},public;").format(Identifier(schema)) + copy_query

    with engine.copy_cursor() as cur:
        cur.copy_expert(copy_query, buffer)


class CSVDataSource(ForeignDataWrapperDataSource):
    credentials_schema: Dict[str, Any] = {
        "type": "object",
        "properties": {
            "s3_access_key": {"type": "string", "title": "AWS Access Key"},
            "s3_secret_key": {"type": "string", "title": "AWS Secret Access Key"},
        },
    }

    params_schema: Dict[str, Any] = {
        "type": "object",
        "properties": {
            "connection": {
                "type": "object",
                "title": "Connection",
                "oneOf": [
                    {
                        "type": "object",
                        "required": ["connection_type", "url"],
                        "properties": {
                            "connection_type": {
                                "type": "string",
                                "const": "http",
                                "title": "Connection type",
                            },
                            "url": {
                                "type": "string",
                                "description": "HTTP URL to the CSV file",
                                "title": "URL",
                            },
                        },
                    },
                    {
                        "type": "object",
                        "required": ["connection_type", "s3_endpoint", "s3_bucket"],
                        "properties": {
                            "connection_type": {
                                "type": "string",
                                "const": "s3",
                                "title": "Connection type",
                            },
                            "s3_endpoint": {
                                "type": "string",
                                "title": "S3 endpoint",
                                "description": "S3 endpoint (including port if required)",
                            },
                            "s3_region": {
                                "type": "string",
                                "title": "S3 region",
                                "description": "Region of the S3 bucket",
                            },
                            "s3_secure": {
                                "title": "Secure",
                                "type": "boolean",
                                "description": "Whether to use HTTPS for S3 access",
                            },
                            "s3_bucket": {
                                "title": "Bucket name",
                                "type": "string",
                                "description": "Bucket the object is in",
                            },
                            "s3_object": {
                                "title": "S3 Object name",
                                "type": "string",
                                "description": "Limit the import to a single object",
                            },
                            "s3_object_prefix": {
                                "title": "S3 Object prefix",
                                "type": "string",
                                "description": "Prefix for object in S3 bucket",
                            },
                        },
                    },
                ],
            },
            "autodetect_header": {
                "type": "boolean",
                "title": "Autodetect header",
                "description": "Detect whether the CSV file has a header automatically",
                "default": True,
            },
            "autodetect_dialect": {
                "type": "boolean",
                "title": "Autodetect dialect",
                "description": "Detect the CSV file's dialect (separator, quoting characters etc) automatically",
                "default": True,
            },
            "autodetect_encoding": {
                "type": "boolean",
                "title": "Autodetect encoding",
                "description": "Detect the CSV file's encoding automatically",
                "default": True,
            },
            "autodetect_sample_size": {
                "type": "integer",
                "title": "Sample size",
                "description": "Sample size, in bytes, for encoding/dialect/header detection",
                "default": 65536,
            },
            "schema_inference_rows": {
                "type": "integer",
                "title": "Schema inference rows",
                "description": "Number of rows to use for schema inference",
                "default": 100000,
            },
            "encoding": {
                "type": "string",
                "title": "Encoding",
                "description": "Encoding of the CSV file",
                "default": "utf-8",
            },
            "ignore_decode_errors": {
                "type": "boolean",
                "title": "Ignore decoding errors",
                "description": "Ignore errors when decoding the file",
                "default": False,
            },
            "header": {
                "type": "boolean",
                "description": "First line of the CSV file is its header",
                "default": True,
            },
            "delimiter": {
                "title": "Delimiter",
                "type": "string",
                "description": "Character used to separate fields in the file",
                "default": ",",
            },
            "quotechar": {
                "title": "Quote character",
                "type": "string",
                "description": "Character used to quote fields",
                "default": '"',
            },
        },
    }

    table_params_schema: Dict[str, Any] = {
        "type": "object",
        "properties": {
            "url": {"type": "string", "description": "HTTP URL to the CSV file", "title": "URL"},
            "s3_object": {
                "type": "string",
                "description": "S3 object of the CSV file",
                "title": "S3 object",
            },
            # Add various CSV dialect overrides to the table params schema.
            **{k: v for k, v in params_schema["properties"].items() if k != "connection"},
        },
    }

    supports_mount = True
    supports_load = True
    supports_sync = False

    _icon_file = "csv.svg"

    commandline_help = """Mount CSV files in S3/HTTP.

If passed an URL, this will live query a CSV file on an HTTP server. If passed
S3 access credentials, this will scan a bucket for CSV files, infer their schema
and make them available to query over SQL.  

For example:  

\b
```
sgr mount csv target_schema -o@- <<EOF
  {
    "s3_endpoint": "cdn.mycompany.com:9000",
    "s3_access_key": "ABCDEF",
    "s3_secret_key": "GHIJKL",
    "s3_bucket": "data",
    "s3_object_prefix": "csv_files/current/",
    "autodetect_header": true,
    "autodetect_dialect": true,
    "autodetect_encoding": true
  }
EOF
```
"""

    commandline_kwargs_help: str = (
        build_commandline_help(credentials_schema) + "\n" + build_commandline_help(params_schema)
    )

    def __init__(
        self,
        engine: "PostgresEngine",
        credentials: Credentials,
        params: Params,
        tables: Optional[TableInfo] = None,
    ):
        # TODO this is a hack to automatically accept both old and new versions of CSV params.
        #  We might need a more robust data source config migration system.
        params = CSVDataSource.migrate_params(params)
        super().__init__(engine, credentials, params, tables)

    def get_fdw_name(self):
        return "multicorn"

    @classmethod
    def get_name(cls) -> str:
        return "CSV files in S3/HTTP"

    @classmethod
    def get_description(cls) -> str:
        return "CSV files in S3/HTTP"

    @classmethod
    def from_commandline(cls, engine, commandline_kwargs) -> "CSVDataSource":
        params = deepcopy(commandline_kwargs)
        credentials = Credentials({})
        for k in ["s3_access_key", "s3_secret_key"]:
            if k in params:
                credentials[k] = params[k]
        return cls(engine, credentials, params)

    @classmethod
    def migrate_params(cls, params: Params) -> Params:
        params = deepcopy(params)
        if "connection" in params:
            return params

        if "url" in params:
            params["connection"] = {"connection_type": "http", "url": params["url"]}
            del params["url"]
        else:
            connection = {"connection_type": "s3"}
            for key in [
                "s3_endpoint",
                "s3_region",
                "s3_secure",
                "s3_bucket",
                "s3_object",
                "s3_object_prefix",
            ]:
                with contextlib.suppress(KeyError):
                    connection[key] = params.pop(key)

            params["connection"] = connection
        return params

    def get_table_options(
        self, table_name: str, tables: Optional[TableInfo] = None
    ) -> Dict[str, str]:
        tables = tables or self.tables

        if not isinstance(tables, dict):
            return {}

        result = {
            k: v
            for k, v in tables.get(table_name, cast(Tuple[TableSchema, TableParams], ({}, {})))[
                1
            ].items()
        }

        # Set a default s3_object if we're using S3 and not HTTP
        if "url" not in result:
            result["s3_object"] = result.get(
                "s3_object", self.params.get("s3_object_prefix", "") + table_name
            )
        return dump_options(result)

    def _create_foreign_tables(
        self, schema: str, server_id: str, tables: TableInfo
    ) -> List[MountError]:
        # Override _create_foreign_tables (actual mounting code) to support TableInfo structs
        # where the schema is empty. This is so that we can call this data source with a limited
        # list of CSV files and their delimiters / other params and have it introspect just
        # those tables (instead of e.g. scanning the whole bucket).
        errors: List[MountError] = []

        if isinstance(tables, dict):
            to_introspect = {
                table_name: table_options
                for table_name, (table_schema, table_options) in tables.items()
                if not table_schema
            }
            if to_introspect:
                # This FDW's implementation of IMPORT FOREIGN SCHEMA supports passing a JSON of table
                # options in options
                errors.extend(
                    import_foreign_schema(
                        self.engine,
                        schema,
                        self.get_remote_schema_name(),
                        server_id,
                        tables=list(to_introspect.keys()),
                        options={"table_options": json.dumps(to_introspect)},
                    )
                )

            # Create the remaining tables (that have a schema) as usual.
            tables = {
                table_name: (table_schema, table_options)
                for table_name, (table_schema, table_options) in tables.items()
                if table_schema
            }
            if not tables:
                return errors

        errors.extend(super()._create_foreign_tables(schema, server_id, tables))
        return errors

    def _get_foreign_table_options(self, schema: str) -> List[Tuple[str, Dict[str, Any]]]:
        options = super()._get_foreign_table_options(schema)

        # Deserialize things like booleans from the foreign table options that the FDW inferred
        # for us.
        return [(t, load_options(d)) for t, d in options]

    def get_server_options(self):
        options: Dict[str, Any] = {}
        for k in self.params_schema["properties"].keys():
            # Flatten the options and extract connection parameters
            if k in self.params:
                if k != "connection":
                    options[k] = self.params[k]
                else:
                    options.update(self.params[k])

        for k in self.credentials_schema["properties"].keys():
            if k in self.credentials:
                options[k] = self.credentials[k]
        # Serialize all options to JSON so that we can be sure they're all strings.
        options = dump_options(options)
        options["wrapper"] = "splitgraph.ingestion.csv.fdw.CSVForeignDataWrapper"

        return options

    def get_remote_schema_name(self) -> str:
        # We ignore the schema name and use the bucket/prefix passed in the params instead.
        return "data"

    @staticmethod
    def _get_url(merged_options: Dict[str, Any], expiry: int = 3600) -> List[Tuple[str, str]]:
        from splitgraph.ingestion.csv.common import get_s3_params

        if merged_options.get("url"):
            # Currently all URLs are public anyway, so return it directly
            return [("text/csv", merged_options["url"])]
        else:
            # Instantiate the Minio client and pre-sign a URL
            s3_client, s3_bucket, s3_object_prefix = get_s3_params(merged_options)
            s3_object = merged_options.get("s3_object")
            if not s3_object:
                return []
            return [
                (
                    "text/csv",
                    s3_client.presigned_get_object(
                        bucket_name=s3_bucket,
                        object_name=s3_object,
                        expires=timedelta(seconds=expiry),
                    ),
                )
            ]

    def get_raw_url(
        self, tables: Optional[TableInfo] = None, expiry: int = 3600
    ) -> Dict[str, List[Tuple[str, str]]]:
        tables = tables or self.tables
        if not tables:
            return {}
        result: Dict[str, List[Tuple[str, str]]] = {}

        # Merge the table options to take care of overrides and use them to get URLs
        # for each table.
        server_options = self.get_server_options()
        for table in tables:
            table_options = self.get_table_options(table, tables)
            full_options = {**server_options, **table_options}
            del full_options["wrapper"]
            result[table] = self._get_url(load_options(full_options))

        return result



"""
Exceptions that can be raised by the Splitgraph library.
"""
from typing import List, Optional


class SplitGraphError(Exception):
    """A generic Splitgraph exception."""


class CheckoutError(SplitGraphError):
    """Errors related to checking out/committing repositories"""


class UnsupportedSQLError(SplitGraphError):
    """Raised for unsupported SQL statements, for example, containing schema-qualified tables when the statement
    is supposed to be used in an SQL/IMPORT Splitfile command."""


class EngineInitializationError(SplitGraphError):
    """Raised when the engine isn't initialized (no splitgraph_meta schema or audit triggers)"""


class EngineSetupError(SplitGraphError):
    """Raised on various setup issues with the Splitgraph engine"""


class ObjectCacheError(SplitGraphError):
    """Issues with the object cache (not enough space)"""


class ObjectNotFoundError(SplitGraphError):
    """Raised when a physical object doesn't exist in the cache."""


class ObjectIndexingError(SplitGraphError):
    """Errors related to indexing objects"""


class ObjectMountingError(SplitGraphError):
    """Errors related to incompatible objects."""


class RepositoryNotFoundError(SplitGraphError):
    """A Splitgraph repository doesn't exist."""


class ImageNotFoundError(SplitGraphError):
    """A Splitgraph image doesn't exist."""


class TableNotFoundError(SplitGraphError):
    """A table doesn't exist in an image"""


class SplitfileError(SplitGraphError):
    """Generic error class for Splitfile interpretation/execution errors."""


class ExternalHandlerError(SplitGraphError):
    """Exceptions raised by external object handlers."""


class DataSourceError(SplitGraphError):
    """Exceptions raised by mount handlers."""


class AuthAPIError(SplitGraphError):
    """Exceptions raised by the Auth API"""


class APICompatibilityError(SplitGraphError):
    """Exceptions related to API incompatibilities"""


class DockerUnavailableError(SplitGraphError):
    """Could not connect to the Docker daemon."""


class IncompleteObjectUploadError(SplitGraphError):
    """Raised when an error is encountered during upload of multiple objects.
    The handler is supposed to perform any necessary
    cleanup and reraise `reason` at the earliest opportunity."""

    def __init__(
        self,
        reason: Optional[BaseException],
        successful_objects: List[str],
        successful_object_urls: List[str],
    ):
        self.reason = reason
        self.successful_objects = successful_objects
        self.successful_object_urls = successful_object_urls


class IncompleteObjectDownloadError(SplitGraphError):
    """Raised when an error is encountered during download of multiple objects.
    The handler is supposed to perform any necessary
    cleanup and reraise `reason` at the earliest opportunity."""

    def __init__(
        self,
        reason: Optional[BaseException],
        successful_objects: List[str],
    ):
        self.reason = reason
        self.successful_objects = successful_objects


class GQLAPIError(SplitGraphError):
    """GQL API errors"""


class GQLUnauthorizedError(GQLAPIError):
    """Unauthorized (e.g. repository isn't writeable by the user)"""


class GQLUnauthenticatedError(GQLAPIError):
    """Unauthenticated (user not logged in)"""


class GQLRepoDoesntExistError(GQLAPIError):
    """Repository doesn't exist"""


class JSONSchemaValidationError(SplitGraphError):
    """Error validating the remote schema"""

    def __init__(self, message: str):
        self.message = message


def get_exception_name(o):
    module = o.__class__.__module__
    if module is None or module == str.__class__.__module__:
        return o.__class__.__name__
    return module + "." + o.__class__.__name__



import base64
import json
import urllib.parse
from copy import deepcopy
from typing import TYPE_CHECKING, Any, Dict, Optional

from splitgraph.core.types import Credentials, Params, TableInfo
from splitgraph.hooks.data_source.fdw import ForeignDataWrapperDataSource
from splitgraph.ingestion.common import build_commandline_help

if TYPE_CHECKING:
    from splitgraph.engine.postgres.engine import PostgresEngine


def _encode_private_key(privkey: str):
    from cryptography.hazmat.backends import default_backend
    from cryptography.hazmat.primitives import serialization

    if "PRIVATE KEY" in privkey:
        # Strip various markers and newlines from the private key, leaving
        # just the b64-encoded body.
        p_key = serialization.load_pem_private_key(
            privkey.strip().encode("ascii"), password=None, backend=default_backend()
        )
        return base64.b64encode(
            p_key.private_bytes(
                encoding=serialization.Encoding.DER,
                format=serialization.PrivateFormat.PKCS8,
                encryption_algorithm=serialization.NoEncryption(),
            )
        ).decode()

    return privkey


class SnowflakeDataSource(ForeignDataWrapperDataSource):
    credentials_schema: Dict[str, Any] = {
        "type": "object",
        "properties": {
            "username": {"type": "string", "title": "Username", "description": "Username"},
            "secret": {
                "type": "object",
                "title": "Connection secret",
                "oneOf": [
                    {
                        "type": "object",
                        "required": ["secret_type", "password"],
                        "properties": {
                            "secret_type": {
                                "type": "string",
                                "const": "password",
                                "title": "Secret type",
                            },
                            "password": {
                                "type": "string",
                                "title": "Password",
                                "description": "Password",
                            },
                        },
                    },
                    {
                        "type": "object",
                        "required": ["secret_type", "private_key"],
                        "properties": {
                            "secret_type": {
                                "type": "string",
                                "const": "private_key",
                                "title": "Secret type",
                            },
                            "private_key": {
                                "type": "string",
                                "title": "Private key",
                                "description": "Private key in PEM format",
                            },
                        },
                    },
                ],
                "description": "A password or a private key to be used for authentication",
            },
            "account": {
                "type": "string",
                "description": "Account Locator, e.g. xy12345.us-east-2.aws. For more information, see https://docs.snowflake.com/en/user-guide/connecting.html",
            },
        },
        "required": ["username", "account"],
    }

    params_schema = {
        "type": "object",
        "properties": {
            "database": {
                "type": "string",
                "title": "Database",
                "description": "Snowflake database name",
            },
            "schema": {"type": "string", "title": "Schema", "description": "Snowflake schema"},
            "warehouse": {"type": "string", "title": "Warehouse", "description": "Warehouse name"},
            "role": {"type": "string", "title": "Role", "description": "Role"},
            "batch_size": {
                "type": "integer",
                "title": "Batch size",
                "description": "Default fetch size for remote queries",
            },
            "envvars": {
                "type": "object",
                "title": "Environment variables",
                "description": "Environment variables to set on the engine side",
            },
        },
        "required": ["database"],
    }

    table_params_schema = {
        "type": "object",
        "properties": {
            "subquery": {
                "type": "string",
                "title": "Subquery",
                "description": "Subquery for this table to run on the server side",
            }
        },
    }
    supports_mount = True
    supports_load = True
    supports_sync = False

    commandline_help = """Mount a Snowflake database.

This will mount a remote Snowflake schema or a table. You can also get a mounted table to point to the result of a subquery that will be executed on the Snowflake instance. For example:

\b
```
$ sgr mount snowflake test_snowflake -o@- <<EOF
{
    "username": "username",
    "secret": {
		"secret_type": "password",
		"password": "password"
	},
    "account": "acc-id.west-europe.azure",
    "database": "SNOWFLAKE_SAMPLE_DATA",
    "schema": "TPCH_SF100",
    "envvars": {"HTTPS_PROXY": "http://proxy.company.com"}
}
EOF
\b
$ sgr mount snowflake test_snowflake_subquery -o@- <<EOF
{
    "username": "username",
    "secret": {
		"secret_type": "private_key",
		"private_key": "MIIEvQIBAD..."
	},
    "account": "acc-id.west-europe.azure",
    "database": "SNOWFLAKE_SAMPLE_DATA",
    "tables": {
        "balances": {
            "schema": {
                "n_nation": "varchar",
                "segment": "varchar",
                "avg_balance": "numeric"
            },
            "options": {
                "subquery": "SELECT n_nation AS nation, c_mktsegment AS segment, AVG(c_acctbal) AS avg_balance FROM TPCH_SF100.customer c JOIN TPCH_SF100.nation n ON c_nationkey = n_nationkey"
            }
        }
    }
}
EOF
```
    """

    commandline_kwargs_help: str = (
        build_commandline_help(credentials_schema)
        + "\n"
        + build_commandline_help(params_schema)
        + "\n"
        + "The schema parameter is required when subquery isn't used."
    )

    _icon_file = "snowflake.svg"

    def __init__(
        self,
        engine: "PostgresEngine",
        credentials: Credentials,
        params: Params,
        tables: Optional[TableInfo] = None,
    ):
        # TODO this is a hack to automatically accept both old and new versions of CSV params.
        #  We might need a more robust data source config migration system.
        credentials = SnowflakeDataSource.migrate_credentials(credentials)
        super().__init__(engine, credentials, params, tables)

    def get_fdw_name(self):
        return "multicorn"

    @classmethod
    def get_name(cls) -> str:
        return "Snowflake"

    @classmethod
    def get_description(cls) -> str:
        return "Schema, table or a subquery from a Snowflake database"

    def get_table_options(
        self, table_name: str, tables: Optional[TableInfo] = None
    ) -> Dict[str, str]:
        result = super().get_table_options(table_name, tables)
        result["tablename"] = result.get("tablename", table_name)
        return result

    def get_server_options(self):
        options: Dict[str, Optional[str]] = {
            "wrapper": "multicorn.sqlalchemyfdw.SqlAlchemyFdw",
            "db_url": self._build_db_url(),
        }

        # For some reason, in SQLAlchemy, if this is not passed
        # to the FDW params (even if it is in the DB URL), it doesn't
        # schema-qualify tables and server-side cursors don't work for scanning
        # (loads the whole table instead of scrolling through it).
        if "schema" in self.params:
            options["schema"] = self.params["schema"]

        if "envvars" in self.params:
            options["envvars"] = json.dumps(self.params["envvars"])

        if "batch_size" in self.params:
            options["batch_size"] = str(self.params["batch_size"])

        if self.credentials["secret"]["secret_type"] == "private_key":
            options["connect_args"] = json.dumps(
                {"private_key": _encode_private_key(self.credentials["secret"]["private_key"])}
            )

        return options

    @classmethod
    def migrate_credentials(cls, credentials: Credentials) -> Credentials:
        credentials = deepcopy(credentials)
        if "private_key" in credentials:
            credentials["secret"] = {
                "secret_type": "private_key",
                "private_key": credentials.pop("private_key"),
            }
        elif "password" in credentials:
            credentials["secret"] = {
                "secret_type": "password",
                "password": credentials.pop("password"),
            }

        return credentials

    def _build_db_url(self) -> str:
        """Construct the SQLAlchemy Snowflake db_url"""

        uname = self.credentials["username"]
        if self.credentials["secret"]["secret_type"] == "password":
            uname += f":{self.credentials['secret']['password']}"

        db_url = f"snowflake://{uname}@{self.credentials['account']}"
        if "database" in self.params:
            db_url += f"/{self.params['database']}"
            if "schema" in self.params:
                db_url += f"/{self.params['schema']}"

        extra_params = {}

        if "warehouse" in self.params:
            extra_params["warehouse"] = self.params["warehouse"]

        if "role" in self.params:
            extra_params["role"] = self.params["role"]

        if extra_params:
            db_url += "?" + urllib.parse.urlencode(extra_params)

        return db_url

    def get_remote_schema_name(self) -> str:
        if "schema" not in self.params:
            raise ValueError("Cannot IMPORT FOREIGN SCHEMA without a schema!")
        return str(self.params["schema"])



"""Routines that are run inside of the engine,
here so that they can get type- and syntax-checked.

When inside of an LQFDW shim, these are called directly by the Splitgraph core code
to avoid a redundant connection to the engine.
"""
import contextlib
import os.path
from typing import List, Tuple
from urllib.parse import urlparse

from splitgraph.config import CONFIG

SG_ENGINE_OBJECT_PATH = str(CONFIG["SG_ENGINE_OBJECT_PATH"])

# An object consists of three files: CStore file, CStore footer and the JSON schema spec.
# We have to download them separately.
ObjectUrls = Tuple[str, str, str]


def verify(url: str):
    # If there's a file called /rootCA.pem in the engine, use it as the CA for
    # HTTPS S3 operations with .test domains (for testing with self-signed certs)
    hostname = urlparse(url).hostname
    if hostname and hostname.split(".")[-1] == "test":
        return "/rootCA.pem" if os.path.exists("/rootCA.pem") else False
    else:
        return True


def _remove(path: str):
    # Wrapper around remove that doesn't throw if the file doesn't exist.
    with contextlib.suppress(FileNotFoundError):
        os.remove(path)


def upload_object(object_id: str, urls: ObjectUrls):
    import requests

    object_path = os.path.join(SG_ENGINE_OBJECT_PATH, object_id)

    for suffix, url in zip(("", ".footer", ".schema"), urls):
        with open(object_path + suffix, "rb") as f:
            response = requests.put(url, data=f, verify=verify(url))
            response.raise_for_status()


def download_object(object_id: str, urls: ObjectUrls):
    import shutil

    import requests

    object_path = os.path.join(SG_ENGINE_OBJECT_PATH, object_id)
    for suffix, url in zip(("", ".footer", ".schema"), urls):
        with requests.get(url, stream=True, verify=verify(url)) as response:
            response.raise_for_status()
            with open(object_path + suffix, "wb") as f:
                shutil.copyfileobj(response.raw, f)


def set_object_schema(object_id: str, schema: str):
    with open(os.path.join(SG_ENGINE_OBJECT_PATH, object_id + ".schema"), "w") as f:
        f.write(schema)


def get_object_schema(object_id: str) -> str:
    with open(os.path.join(SG_ENGINE_OBJECT_PATH, object_id + ".schema")) as f:
        return f.read()


def delete_object_files(object_id: str):
    object_path = os.path.join(SG_ENGINE_OBJECT_PATH, object_id)
    _remove(object_path)
    _remove(object_path + ".footer")
    _remove(object_path + ".schema")


def rename_object_files(old_object_id: str, new_object_id: str):
    import shutil

    for suffix in ("", ".footer", ".schema"):
        shutil.move(
            os.path.join(SG_ENGINE_OBJECT_PATH, old_object_id + suffix),
            os.path.join(SG_ENGINE_OBJECT_PATH, new_object_id + suffix),
        )


def get_object_size(object_id: str) -> int:
    object_path = os.path.join(SG_ENGINE_OBJECT_PATH, object_id)
    return (
        os.path.getsize(object_path)
        + os.path.getsize(object_path + ".footer")
        + os.path.getsize(object_path + ".schema")
    )


def list_objects() -> List[str]:
    from collections import defaultdict

    # Crude but faster than listing foreign tables (and hopefully consistent).
    files = os.listdir(SG_ENGINE_OBJECT_PATH)

    # Make sure to only return objects that have been fully downloaded.
    objects = defaultdict(list)
    for f in files:
        objects[f.replace(".schema", "").replace(".footer", "")].append(f)

    return [f for f, fs in objects.items() if len(fs) == 3]


def object_exists(object_id: str) -> bool:
    # Check if the physical object file exists in storage.
    # Make sure to check for all 3 files to guard against partially failed writes.
    return all(
        os.path.exists(os.path.join(SG_ENGINE_OBJECT_PATH, object_id + suffix))
        for suffix in ("", ".footer", ".schema")
    )



import os
import tarfile
import time
from io import BytesIO
from tarfile import TarFile, TarInfo
from typing import IO, TYPE_CHECKING, List, Optional

if TYPE_CHECKING:
    from docker.models.containers import Container

from splitgraph.exceptions import DockerUnavailableError


def get_docker_client():
    """Wrapper around client.from_env() that also pings the daemon
    to make sure it can connect and if not, raises an error."""
    import docker

    try:
        client = docker.from_env()
        client.ping()
        return client
    except Exception as e:
        raise DockerUnavailableError("Could not connect to the Docker daemon") from e


def copy_to_container(
    container: "Container",
    source_path: Optional[str],
    target_path: str,
    data: Optional[bytes] = None,
) -> None:
    """
    Copy a file into a Docker container

    :param container: Container object
    :param source_path: Source file path
    :param target_path: Target file path (in the container)
    :return:
    """

    if data is None:
        if not source_path:
            raise ValueError("One of source_path or data must be specified!")
        # https://github.com/docker/docker-py/issues/1771
        with open(source_path, "rb") as f:
            data = f.read()

    tarinfo = TarInfo(name=os.path.basename(target_path))
    tarinfo.size = len(data)
    tarinfo.mtime = int(time.time())

    stream = BytesIO()
    tar = TarFile(fileobj=stream, mode="w")
    tar.addfile(tarinfo, BytesIO(data))
    tar.close()

    stream.seek(0)
    container.put_archive(path=os.path.dirname(target_path), data=stream.read())


def copy_dir_to_container(
    container: "Container",
    source_path: str,
    target_path: str,
    exclude_names: Optional[List[str]] = None,
) -> None:
    exclude_names = exclude_names or []
    stream = BytesIO()
    tar = TarFile(fileobj=stream, mode="w")
    tar.add(
        name=source_path,
        arcname=target_path,
        recursive=True,
        filter=lambda ti: None if ti.name in exclude_names else ti,
    )
    tar.close()

    stream.seek(0)
    container.put_archive(path="/", data=stream.read())


def get_file_from_container(
    container: "Container",
    source_path: str,
) -> IO[bytes]:
    archive, stat = container.get_archive(source_path)
    file_obj = BytesIO()
    for chunk in archive:
        file_obj.write(chunk)
    file_obj.seek(0)
    with tarfile.open(mode="r", fileobj=file_obj) as tar:
        data = tar.extractfile(stat["name"])
        assert data
        return data



import gzip
import json
import logging
import os
from contextlib import contextmanager
from copy import copy
from itertools import islice
from typing import Optional

import requests
from urllib3 import HTTPResponse

import splitgraph.config
from splitgraph.config import get_singleton
from splitgraph.exceptions import get_exception_name
from splitgraph.ingestion.common import generate_column_names
from splitgraph.ingestion.csv.common import (
    CSVOptions,
    dump_options,
    get_s3_params,
    load_options,
    make_csv_reader,
    pad_csv_row,
)
from splitgraph.ingestion.inference import infer_sg_schema

try:
    from multicorn import ANY, ColumnDefinition, ForeignDataWrapper, TableDefinition
except ImportError:
    # Multicorn not installed (OK if we're not on the engine -- tests).
    ForeignDataWrapper = object
    ANY = object()
    TableDefinition = dict
    ColumnDefinition = dict

try:
    from multicorn.utils import log_to_postgres
except ImportError:

    def log_to_postgres(*args, **kwargs):
        print(*args)


_PG_LOGLEVEL = logging.INFO


def _get_table_definition(response, fdw_options, table_name, table_options):
    # Allow overriding introspection options with per-table params (e.g. encoding, delimiter...)
    fdw_options = copy(fdw_options)
    fdw_options.update(table_options)

    csv_options, reader = make_csv_reader(response, CSVOptions.from_fdw_options(fdw_options))
    sample = list(islice(reader, csv_options.schema_inference_rows))

    if not csv_options.header:
        sample = [[""] * len(sample[0])] + sample

    # Ignore empty lines (newlines at the end of file etc)
    sample = [row for row in sample if len(row) > 0]

    sg_schema = infer_sg_schema(sample, None, None)

    # For nonexistent column names: replace with autogenerated ones (can't have empty column names)
    sg_schema = generate_column_names(sg_schema)

    # Merge the autodetected table options with the ones passed to us originally (e.g.
    # S3 object etc)
    new_table_options = copy(table_options)
    new_table_options.update(csv_options.to_table_options())

    # Build Multicorn TableDefinition. ColumnDefinition takes in type OIDs,
    # typmods and other internal PG stuff but other FDWs seem to get by with just
    # the textual type name.
    return TableDefinition(
        table_name=table_name,
        schema=None,
        columns=[ColumnDefinition(column_name=c.name, type_name=c.pg_type) for c in sg_schema],
        options=dump_options(new_table_options),
    )


@contextmanager
def report_errors(table_name: str):
    """Context manager that ignores exceptions and serializes them to JSON using PG's notice
    mechanism instead. The data source is meant to load these to report on partial failures
    (e.g. failed to load one table, but not others)."""
    try:
        yield
    except Exception as e:
        logging.error(
            "Error scanning %s, ignoring: %s: %s",
            table_name,
            get_exception_name(e),
            e,
            exc_info=e,
        )
        log_to_postgres(
            "SPLITGRAPH: "
            + json.dumps(
                {
                    "table_name": table_name,
                    "error": get_exception_name(e),
                    "error_text": str(e),
                }
            )
        )


class CSVForeignDataWrapper(ForeignDataWrapper):
    """Foreign data wrapper for CSV files stored in S3 buckets or HTTP"""

    def __init__(self, fdw_options, fdw_columns):
        # Initialize the logger that will log to the engine's stderr: log timestamp and PID.

        logging.basicConfig(
            format="%(asctime)s [%(process)d] %(levelname)s %(message)s",
            level=get_singleton(splitgraph.config.CONFIG, "SG_LOGLEVEL"),
        )

        # Dict of connection parameters
        self.fdw_options = load_options(fdw_options)

        # The foreign datawrapper columns (name -> ColumnDefinition).
        self.fdw_columns = fdw_columns
        self._num_cols = len(fdw_columns)

        self.csv_options = CSVOptions.from_fdw_options(self.fdw_options)

        # For HTTP: use full URL
        if self.fdw_options.get("url"):
            self.mode = "http"
            self.url = self.fdw_options["url"]
        else:
            self.mode = "s3"
            self.s3_client, self.s3_bucket, self.s3_object_prefix = get_s3_params(self.fdw_options)

            self.s3_object = self.fdw_options["s3_object"]

    def can_sort(self, sortkeys):
        # Currently, can't sort on anything. In the future, we can infer which
        # columns a CSV is sorted on and return something more useful here.
        return []

    def get_rel_size(self, quals, columns):
        return 1000000, len(columns) * 10

    def explain(self, quals, columns, sortkeys=None, verbose=False):
        if self.mode == "http":
            return ["HTTP request", f"URL: {self.url}"]
        else:
            return [
                "S3 request",
                f"Endpoint: {self.s3_client._base_url}",
                f"Bucket: {self.s3_bucket}",
                f"Object ID: {self.s3_object}",
            ]

    def _read_csv(self, csv_reader, csv_options):
        header_skipped = False
        for row_number, row in enumerate(csv_reader):
            if not header_skipped and csv_options.header:
                header_skipped = True
                continue

            # Ignore empty rows too
            if not row:
                continue

            row = pad_csv_row(row, row_number=row_number, num_cols=self._num_cols)

            # CSVs don't really distinguish NULLs and empty strings well. We know
            # that empty strings should be NULLs when coerced into non-strings but we
            # can't easily access type information here. Do a minor hack and treat
            # all empty strings as NULLs.
            row = [r if r != "" else None for r in row]

            yield row

    def execute(self, quals, columns, sortkeys=None):
        """Main Multicorn entry point."""

        if self.mode == "http":
            with requests.get(
                self.url, stream=True, verify=os.environ.get("SSL_CERT_FILE", True)
            ) as response:
                response.raise_for_status()
                stream = response.raw
                if response.headers.get("Content-Encoding") == "gzip":
                    stream = gzip.GzipFile(fileobj=stream)

                csv_options = self.csv_options
                if csv_options.encoding == "" and not csv_options.autodetect_encoding:
                    csv_options = csv_options._replace(encoding=response.encoding)

                csv_options, reader = make_csv_reader(stream, csv_options)
                yield from self._read_csv(reader, csv_options)
        else:
            minio_response: Optional[HTTPResponse] = None
            try:
                minio_response = self.s3_client.get_object(
                    bucket_name=self.s3_bucket, object_name=self.s3_object
                )
                assert minio_response
                csv_options = self.csv_options
                if csv_options.encoding == "" and not csv_options.autodetect_encoding:
                    csv_options = csv_options._replace(autodetect_encoding=True)
                csv_options, reader = make_csv_reader(minio_response, csv_options)
                yield from self._read_csv(reader, csv_options)
            finally:
                if minio_response:
                    minio_response.close()
                    minio_response.release_conn()

    @classmethod
    def import_schema(cls, schema, srv_options, options, restriction_type, restricts):
        # Implement IMPORT FOREIGN SCHEMA to instead scan an S3 bucket for CSV files
        # and infer their CSV schema.

        # 1) if we don't have options["table_options"], do a full scan as normal,
        #    treat LIMIT TO as a list of S3 objects
        # 2) if we do, go through these tables and treat each one as a partial override
        #    of server options
        if "table_options" in options:
            table_options = json.loads(options["table_options"])
        else:
            table_options = None
        srv_options = load_options(srv_options)

        if not table_options:
            # Do a full scan of the file at URL / S3 bucket w. prefix
            if srv_options.get("url"):
                # Infer from HTTP -- singular table with name "data"
                result = cls._introspect_url(srv_options, srv_options["url"])
                if result:
                    return [result]
                return []
            else:
                # Get S3 options
                client, bucket, prefix = get_s3_params(srv_options)

                # Note that we ignore the "schema" here (e.g. IMPORT FOREIGN SCHEMA some_schema)
                # and take all interesting parameters through FDW options.

                # Allow just introspecting one object
                if "s3_object" in srv_options:
                    objects = [srv_options["s3_object"]]
                elif restriction_type == "limit":
                    objects = restricts
                else:
                    objects = [
                        o.object_name
                        for o in client.list_objects(
                            bucket_name=bucket, prefix=prefix or None, recursive=True
                        )
                    ]

                result = []

                for o in objects:
                    if restriction_type == "except" and o in restricts:
                        continue
                    result.append(cls._introspect_s3(client, bucket, o, srv_options))

                return [r for r in result if r]
        else:
            result = []

            # Note we ignore LIMIT/EXCEPT here. There's no point in using them if the user
            # is passing a dict of table options anyway.
            for table_name, this_table_options in table_options.items():
                if "s3_object" in this_table_options:
                    # TODO: we can support overriding S3 params per-table here, but currently
                    #   we don't do it.
                    client, bucket, _ = get_s3_params(srv_options)
                    result.append(
                        cls._introspect_s3(
                            client,
                            bucket,
                            this_table_options["s3_object"],
                            srv_options,
                            table_name,
                            this_table_options,
                        )
                    )
                else:
                    result.append(
                        cls._introspect_url(
                            srv_options, this_table_options["url"], table_name, this_table_options
                        )
                    )
            return [r for r in result if r]

    @classmethod
    def _introspect_s3(
        cls, client, bucket, object_id, srv_options, table_name=None, table_options=None
    ) -> Optional[TableDefinition]:
        response = None
        # Default table name: truncate S3 object key up to the prefix
        table_name = table_name or object_id[len(srv_options.get("s3_object_prefix", "")) :]
        table_options = table_options or {}
        table_options.update({"s3_object": object_id})
        with report_errors(table_name):
            try:
                response = client.get_object(bucket, object_id)
                return _get_table_definition(
                    response,
                    srv_options,
                    table_name,
                    table_options,
                )
            finally:
                if response:
                    response.close()
                    response.release_conn()

    @classmethod
    def _introspect_url(
        cls, srv_options, url, table_name=None, table_options=None
    ) -> Optional[TableDefinition]:
        table_name = table_name or "data"
        table_options = table_options or {}

        with report_errors(table_name):
            with requests.get(
                url, stream=True, verify=os.environ.get("SSL_CERT_FILE", True)
            ) as response:
                response.raise_for_status()
                stream = response.raw
                if response.headers.get("Content-Encoding") == "gzip":
                    stream = gzip.GzipFile(fileobj=stream)
                return _get_table_definition(stream, srv_options, table_name, table_options)



import logging
import os
import re
from contextlib import contextmanager
from typing import Any, List, Tuple

import docker.errors
from docker.models.containers import Container

from splitgraph.exceptions import SplitGraphError
from splitgraph.utils.docker import copy_to_container


class SubprocessError(SplitGraphError):
    pass


def add_files(container: Container, files: List[Tuple[str, str]]) -> None:
    for var_name, var_data in files:
        if not var_data:
            continue
        copy_to_container(
            container,
            source_path=None,
            target_path=f"/{var_name}.json",
            data=var_data.encode(),
        )


@contextmanager
def remove_at_end(container: Container) -> Container:
    try:
        yield container
    finally:
        try:
            container.remove(force=True)
        except docker.errors.APIError as e:
            logging.warning("Error removing container at the end, continuing", exc_info=e)


def wait_not_failed(container: Container, mirror_logs: bool = False) -> None:
    """
    Block until a Docker container exits.

    :raises SubprocessError if the container exited with a non-zero code.
    """

    if mirror_logs:
        for line in container.logs(stream=True, follow=True):
            logging.info("%s: %s", container.name, line.decode().strip())

    result = container.wait()
    if result["StatusCode"] != 0:
        logging.error("Container %s exited with %d", container.name, result["StatusCode"])
        logs = container.logs(tail=1000) or b""
        for line in logs.decode().splitlines():
            logging.info("%s: %s", container.name, line)
        raise SubprocessError(
            "Container %s exited with %d" % (container.name, result["StatusCode"])
        )


def build_command(files: List[Tuple[str, Any]]) -> List[str]:
    command: List[str] = []

    for var_name, var_data in files:
        if not var_data:
            continue
        command.extend([f"--{var_name}", f"/{var_name}.json"])
    return command


def detect_network_mode() -> str:
    # We want the receiver to connect to the same engine that we're connected to. If we're
    # running on the host, that means using our own connection parameters and running the
    # receiver with net:host. Inside Docker we have to use the host's Docker socket and
    # attach the container to our own network so that it can also use our own params.

    # This also applies in case we're running a source against a database that's also running
    # in Docker -- we want to mimic sgr too.
    if os.path.exists("/.dockerenv"):
        with open("/proc/1/cgroup", "r") as f:
            match = re.search(r"^.*/docker/([0-9a-f]{64})$", f.read(), re.MULTILINE)
            if not match:
                raise AssertionError("Could not detect Docker container ID")
        return f"container:{match.group(1)}"
    else:
        return "host"



"""
Routines for rendering a Splitgraph repository as a tree of images
"""
from collections import OrderedDict, defaultdict
from datetime import datetime
from typing import TYPE_CHECKING, List

import click

from splitgraph.config import SG_CMD_ASCII
from splitgraph.core.output import Color

if TYPE_CHECKING:
    from splitgraph.core.repository import Repository


def _pull_up_children(tree):
    """Pull up nodes with only one child to avoid ladders for long
    chains of commits on one branch. Technically, asciitree draws vertical trees,
    but we try to flatten them a bit."""

    # TODO this creates ambiguous trees where a line of commits renders the same
    #   as a bunch of siblings of a single node.
    new_tree: OrderedDict[str, OrderedDict] = OrderedDict()
    one_child = len(tree) == 1

    for child_name, child in tree.items():
        flatten = len(child) == 1
        child = _pull_up_children(child)

        # If this node has a single child that itself only had a single child before
        # flattening, we can pull that tree's new children up (straighten the branch).
        if flatten and one_child:
            new_tree[child_name] = OrderedDict()
            new_tree.update(child)
        else:
            new_tree[child_name] = child

    return new_tree


def format_image_hash(image_hash: str) -> str:
    return Color.BOLD + Color.RED + image_hash[:10] + Color.END


def format_tags(tags: List[str]) -> str:
    if tags:
        return Color.BOLD + Color.YELLOW + " [" + ", ".join(tags) + "]" + Color.END
    return ""


def format_time(time: datetime) -> str:
    return " " + Color.GREEN + time.strftime("%Y-%m-%d %H:%M:%S") + Color.END


def render_tree(repository: "Repository") -> None:
    """Draws the repository's commit graph as a Git-like tree."""

    import asciitree

    from splitgraph.core.output import truncate_line

    # Get all commits in ascending time order
    all_images = {i.image_hash: i for i in repository.images()}

    if not all_images:
        return

    latest = repository.images["latest"]

    tag_dict = defaultdict(list)
    for img, img_tag in repository.get_all_hashes_tags():
        tag_dict[img].append(img_tag)
    tag_dict[latest.image_hash].append("latest")

    class ImageTraversal(asciitree.DictTraversal):
        def get_text(self, node):
            image = all_images[node[0]]
            result = format_image_hash(image.image_hash)
            result += format_tags(tag_dict[image.image_hash])
            result += format_time(image.created)
            if image.comment:
                result += " " + truncate_line(image.comment)
            return result

    tree: OrderedDict[str, OrderedDict] = OrderedDict(
        (image, OrderedDict()) for image in all_images
    )
    tree_elements = tree.copy()

    # Join children to parents to prepare a tree structure for asciitree
    for image in all_images.values():
        if image.parent_id is None or image.parent_id not in tree_elements:
            # If we only pulled a single image, it's possible that we won't have
            # the metadata for the image's parent.
            continue

        tree_elements[image.parent_id][image.image_hash] = tree_elements[image.image_hash]
        del tree[image.image_hash]

    # tree = _pull_up_children(tree)

    renderer = asciitree.LeftAligned(
        draw=asciitree.BoxStyle(
            gfx=asciitree.drawing.BOX_ASCII if SG_CMD_ASCII else asciitree.drawing.BOX_LIGHT,
            label_space=1,
            horiz_len=0,
        ),
        traverse=ImageTraversal(),
    )
    for root, root_tree in tree.items():
        click.echo(renderer({root: root_tree}))



"""
    A framework for custom Splitfile commands. The execution flow is as follows:

      * When the Splitfile executor finds an unknown command, it looks for an entry in the config file::

          [commands]
          RUN=splitgraph.plugins.Run

      * The command class must extend this class, initialized at every invocation time.
      * The command's `calc_hash()` method is run. The resultant command context hash is combined with the current
        image hash to produce the new image hash: if it already exists, then the image is simply checked out.
      * Otherwise (or if calc_hash is undefined or returns None), `execute()`, where the actual command should be
        implemented, is run. If it returns a hash, this hash is used for the new image. If this hash already exists,
        the existing image is checked out instead. If the command returns None, a random hash is generated for the
        new image.
"""


class PluginCommand:
    """Base class for custom Splitfile commands."""

    def calc_hash(self, repository, args):
        """
        Calculates the command context hash for this custom command. If either the command context hash or the
        previous image hash has changed, then the image hash produced by this command will change.
        Consequently, two commands with the same command context hashes are assumed to have the same effect
        on any Splitgraph images.

        This is supposed to be a lightweight method intended for pre-flight image hash calculations
        (without performing the actual transformation). If it returns None, the actual transformation is run anyway.

        For example, for a command that imports some data from an external URL, this could be the hash of the last
        modified timestamp provided by the external data vendor. If the timestamp is unchanged, the data is unchanged
        and so actual command won't be re-executed.

        :param repository: SG Repository object pointed to a schema with the checked out image
            the command is being run against.
        :param args: Positional arguments to the command

        :return: Command context hash (a string of 64 hexadecimal digits)
        """

    def execute(self, repository, args):
        """
        Execute the custom command against the target schema, optionally returning the new image hash. The contract
        for the command is as follows (though it is not currently enforced by the runtime):

          * Has to use get_engine().run_sql (or run_sql_batch) to interact with the engine.
          * Can only write to the schema with the checked-out repository (run_sql runs non-schema-qualified
            statements against the correct schema).
          * Can inspect splitgraph_meta (e.g. to find the current HEAD) for the repository.
          * Can't alter the versioning of the repository.

        :param repository: SG Repository object pointed to a schema with the checked out image
            the command is being run against.
        :param args: Positional arguments to the command

        :return: Command context hash (a string of 64 hexadecimal digits). If calc_hash() had previously returned
            a hash, this hash is ignored. If both this command and calc_hash() return None, the hash is randomly
            generated.
        """



# engine params
from typing import Any, Callable, Dict, Optional, Sequence, Union, cast

from .argument_config import get_argument_config_value
from .config_file_config import get_config_dict_from_config_file
from .default_config import get_default_config_value
from .environment_config import get_environment_config_value
from .keys import ALL_KEYS, KEYS, ConfigDict
from .system_config import get_system_config_value


def chain_getters(
    getters: Sequence[Callable[[str], Optional[str]]],
    key: str,
    default_return: Optional[str] = None,
) -> Optional[str]:
    for getter in getters:
        result = getter(key)
        if result is not None:
            return result
    return default_return


def lazy_get_config_value(
    key: str, default_return: Optional[str] = None
) -> Optional[Union[str, Dict[str, Dict[str, str]]]]:
    """
    Get the config value for a key in the following precedence
    Otherwise return default_return
    """

    if key not in ALL_KEYS:
        # For sections which can't be overridden via envvars/arguments,
        # we only use default values
        return chain_getters([get_default_config_value], key, default_return)

    return chain_getters(
        [
            get_argument_config_value,
            get_environment_config_value,
            get_system_config_value,
            get_default_config_value,
        ],
        key,
        default_return,
    )


def update_config_dict_from_arguments(config_dict: ConfigDict) -> ConfigDict:
    """
    Given an existing config_dict, update after reading sys.argv
    and overwriting any keys.

    Return updated copy of config_dict.
    """
    argument_config_dict = {
        k: get_argument_config_value(k, None)
        for k in KEYS
        if get_argument_config_value(k) is not None
    }
    new_config_dict = patch_config(config_dict, cast(ConfigDict, argument_config_dict))
    return new_config_dict


def update_config_dict_from_env_vars(config_dict: ConfigDict) -> ConfigDict:
    """
    Given an existing config_dict, update after reading os.environ
    and overwriting any keys.

    Return updated copy of config_dict.
    """

    argument_config_dict = {
        k: get_environment_config_value(k, None)
        for k in KEYS
        if get_environment_config_value(k) is not None
    }
    new_config_dict = patch_config(config_dict, cast(ConfigDict, argument_config_dict))

    return new_config_dict


def update_config_dict_from_file(config_dict: ConfigDict, sg_config_file: str) -> ConfigDict:
    """
    Given an existing config_dict, update after reading sg_config_file
    and overwriting any keys according to the rules in config_file_config

    Return updated copy of config_dict.
    """

    config_file_dict = get_config_dict_from_config_file(sg_config_file)
    new_config_dict = patch_config(config_dict, config_file_dict)

    return new_config_dict


def create_config_dict() -> ConfigDict:
    """
    Create and return a dict of all known config values
    """
    initial_dict = {k: lazy_get_config_value(k) for k in ALL_KEYS}
    config_dict = cast(ConfigDict, {k: v for k, v in initial_dict.items() if v is not None})
    try:
        sg_config_file = get_singleton(config_dict, "SG_CONFIG_FILE")
        config_dict = update_config_dict_from_file(config_dict, sg_config_file)
    except KeyError:
        pass
    config_dict = update_config_dict_from_env_vars(config_dict)
    config_dict = update_config_dict_from_arguments(config_dict)

    return config_dict


def patch_config(config: ConfigDict, patch: ConfigDict) -> ConfigDict:
    """
    Recursively updates a nested configuration dictionary:

    patch_config(
        {"key_1": "value_1",
         "dict_1": {"key_1": "value_1"}},
        {"key_1": "value_2",
         "dict_1": {"key_2": "value_2"}}) == \
        {"key_1": "value_2",
         "dict_1": {"key_1": "value_1", "key_2": "value_2"}}

    :param config: Config dictionary
    :param patch: Dictionary with the path
    :return: New patched dictionary
    """

    def _patch_internal(left: Dict[str, Any], right: Dict[str, Any]) -> Dict[str, Any]:
        result = left.copy()
        for key, value in right.items():
            if key in left and isinstance(left[key], dict) and isinstance(value, dict):
                result[key] = _patch_internal(left[key], value)
            else:
                result[key] = value
        return result

    return _patch_internal(config, patch)


def get_singleton(config: ConfigDict, item: str) -> str:
    """Return a singleton (not a section) variable from the config."""
    return str(config[item])


def get_all_in_section(config: ConfigDict, section: str) -> Dict[str, Union[str, Dict[str, str]]]:
    """
    Get all subsections from a config (e.g. config["data_sources"])
    """
    result: Dict[str, Union[str, Dict[str, str]]] = cast(
        Dict[str, Union[str, Dict[str, str]]], config.get(section, {})
    )
    assert isinstance(result, dict)
    return result


def get_all_in_subsection(config: ConfigDict, section: str, subsection: str) -> Dict[str, str]:
    section_dict = get_all_in_section(config, section)
    subsection_dict: Dict[str, str] = cast(Dict[str, str], section_dict.get(subsection, {}))
    assert isinstance(subsection_dict, dict)
    return subsection_dict


def get_from_subsection(config: ConfigDict, section: str, subsection: str, item: str) -> str:
    """Return a singleton variable from a subsection of the config,
    e.g. config["remotes"]["data.splitgraph.com"]["SG_ENGINE_HOST"]"""
    subsection_dict = get_all_in_subsection(config, section, subsection)
    return subsection_dict[item]


def get_from_section(config: ConfigDict, section: str, item: str) -> str:
    section_dict = get_all_in_section(config, section)
    assert isinstance(section_dict, dict)
    return cast(str, section_dict[item])


def set_in_subsection(
    config: ConfigDict, section: str, subsection: str, item: str, value: str
) -> None:
    """Set a singleton variable in a subsection of the config,
    e.g. config["remotes"]["data.splitgraph.com"]["SG_ENGINE_HOST"]"""
    subsection_dict = get_all_in_subsection(config, section, subsection)
    subsection_dict[item] = value



"""Module imported by Multicorn on the Splitgraph engine server: a foreign data wrapper that implements
layered querying (read-only queries to Splitgraph tables without materialization)."""
import contextlib
import logging
from typing import TYPE_CHECKING, Callable, Optional

import splitgraph.config
from splitgraph.config import get_singleton
from splitgraph.core.object_manager import ObjectManager
from splitgraph.core.output import pretty_size
from splitgraph.core.repository import Repository, get_engine
from splitgraph.core.table import QueryPlan, Table

if TYPE_CHECKING:
    from splitgraph.engine.postgres.engine import PostgresEngine


with contextlib.suppress(ImportError):
    # Multicorn not installed (OK if we're not on the engine machine).
    from multicorn import ANY, ForeignDataWrapper

_PG_LOGLEVEL = logging.INFO


class QueryingForeignDataWrapper(ForeignDataWrapper):
    """The actual Multicorn LQ FDW class"""

    # Allow injecting custom object manager classes
    object_manager_class = ObjectManager

    def __init__(self, fdw_options, fdw_columns):
        """The foreign data wrapper is initialized on the first query.
        Args:
            fdw_options (dict): The foreign data wrapper options. It is a dictionary
                mapping keys from the sql "CREATE FOREIGN TABLE"
                statement options. It is left to the implementor
                to decide what should be put in those options, and what
                to do with them.

        """
        # Flip the global flag for engine constructors to use (see comment in splitgraph.config).
        splitgraph.config.IN_FDW = True

        # Initialize the logger that will log to the engine's stderr: log timestamp and PID.
        logging.basicConfig(
            format="%(asctime)s [%(process)d] %(levelname)s %(message)s",
            level=get_singleton(splitgraph.config.CONFIG, "SG_LOGLEVEL"),
        )

        # Dict of connection parameters as well as the table, repository and image hash to query.
        self.fdw_options = fdw_options

        # The foreign datawrapper columns (name -> ColumnDefinition).
        self.fdw_columns = fdw_columns

        self._initialize_engines()

        repository = Repository(
            fdw_options["namespace"],
            self.fdw_options["repository"],
            engine=self.engine,
            object_engine=self.object_engine,
            object_manager=self.object_manager_class(
                object_engine=self.object_engine, metadata_engine=self.engine
            ),
        )
        self.table: Table = repository.images[self.fdw_options["image_hash"]].get_table(
            self.fdw_options["table"]
        )

        # Callback that we have to call when end_scan() is called (release objects and temporary tables).
        self.end_scan_callback: Optional[Callable] = None

        # A QueryPlan object for the last query with stats
        self.plan: Optional[QueryPlan] = None

    @staticmethod
    def _quals_to_cnf(quals):
        def _qual_to_cnf(qual):
            if qual.is_list_operator:
                if any(v is None for v in qual.value):
                    # We don't keep track of NULLs so if we get a clause of type
                    # col IN (1,2,3,NULL) we have to look at all objects to see if they contain a NULL.
                    return [[]]
                if qual.list_any_or_all == ANY:
                    # Convert col op ANY(ARRAY[a,b,c...]) into (col op a) OR (col op b)...
                    # which is one single AND clause of multiple ORs
                    return [[(qual.field_name, qual.operator[0], v) for v in qual.value]]
                # Convert col op ALL(ARRAY[a,b,c...]) into (cop op a) AND (col op b)...
                # which is multiple AND clauses of one OR each
                return [[(qual.field_name, qual.operator[0], v)] for v in qual.value]

            if qual.value is None:
                return [[]]
            return [[(qual.field_name, qual.operator, qual.value)]]

        return [q for qual in quals for q in _qual_to_cnf(qual) if q != []]

    def get_rel_size(self, quals, columns):
        """
        Method called from the planner to estimate the resulting relation
        size for a scan.
        It will help the planner in deciding between different types of plans,
        according to their costs.
        Args:
            quals (list): A list of Qual instances describing the filters
                applied to this scan.
            columns (list): The list of columns that must be returned.
        Returns:
            A tuple of the form (expected_number_of_rows, avg_row_width (in bytes))
        """
        cnf_quals = self._quals_to_cnf(quals)
        plan = self.table.get_query_plan(cnf_quals, columns)

        # Estimate the number of rows -- several precision levels here:
        #   * Number of rows in the actual fragments (using metadata -- no need to download
        #   anything) <- you are here
        #   * calling EXPLAIN on all fragments in filtered_objects (might be pretty expensive
        #     and requires the actual fragments to be present)
        #   * reading binary cstore files?
        return plan.estimated_rows, int(
            plan.size_per_row / len(self.table.table_schema) * len(columns)
        )

    def explain(self, quals, columns, sortkeys=None, verbose=False):
        cnf_quals = self._quals_to_cnf(quals)
        plan = self.table.get_query_plan(cnf_quals, columns)
        all_objects = plan.required_objects
        filtered_objects = plan.filtered_objects

        total_size = sum(
            o.size for o in self.table.repository.objects.get_object_meta(filtered_objects).values()
        )

        return [
            "Original Multicorn quals: %r" % quals,
            "CNF quals: %r" % cnf_quals,
            "Objects removed by filter: %d" % (len(all_objects) - len(filtered_objects)),
            "Scan through %d object(s) (%s)" % (len(filtered_objects), pretty_size(total_size)),
        ]

    def get_path_keys(self):
        # Return the PK of the table (unique path something)
        pks = [k[1] for k in self.table.table_schema if k[3]]
        if not pks:
            pks = [k[1] for k in self.table.table_schema]
        return [(tuple(pks), 1)]

    def execute(self, quals, columns, sortkeys=None):
        """Main Multicorn entry point."""
        # We assume that columns here is a list of columns (rather than a set)
        # For quals, the more elaborate ones (like table.id = table.name or similar) actually aren't passed here
        # at all and PG filters them out later on.
        cnf_quals = self._quals_to_cnf(quals)

        # Sometimes (e.g. with running a join) Postgres can ask us to "restart" the scan, which means
        # that it doesn't call our end_scan() method. If we have a scan in progress, make sure to
        # call the callback anyway, since otherwise we'll leak temporary tables.
        if self.end_scan_callback:
            self.end_scan_callback(from_fdw=True)
            self.end_scan_callback = None

        queries, self.end_scan_callback, self.plan = self.table.query_indirect(columns, cnf_quals)
        yield from queries

    def end_scan(self):
        if self.end_scan_callback:
            # Call the scan-end callback making sure to use
            # the special hack to avoid deadlocks with Multicorn.
            self.end_scan_callback(from_fdw=True)
            self.end_scan_callback = None

        self.engine.close()
        self.object_engine.close()

    def _initialize_engines(self):
        # Try using a UNIX socket if the engine is local to us
        self.engine: "PostgresEngine" = get_engine(
            self.fdw_options["engine"],
            bool(self.fdw_options.get("use_socket", False)),
            use_fdw_params=True,
        )
        if "object_engine" in self.fdw_options:
            self.object_engine: "PostgresEngine" = get_engine(
                self.fdw_options["object_engine"],
                bool(self.fdw_options.get("use_socket", False)),
                use_fdw_params=True,
            )
        else:
            self.object_engine = self.engine
